from __future__ import annotations

import pyarrow as pa
import pytest

import daft
from daft import col
from daft.datatype import DataType
from tests.conftest import get_tests_daft_runner_name
from tests.utils import sort_arrow_table


def skip_invalid_join_strategies(join_strategy, join_type):
    if get_tests_daft_runner_name() == "native":
        if join_strategy not in [None, "hash"]:
            pytest.skip("Native executor fails for these tests")
    else:
        if (join_strategy == "sort_merge" or join_strategy == "sort_merge_aligned_boundaries") and join_type != "inner":
            pytest.skip("Sort merge currently only supports inner joins")
        elif join_strategy == "broadcast" and join_type == "outer":
            pytest.skip("Broadcast join does not support outer joins")


def test_invalid_join_strategies(make_df):
    df = make_df(
        {
            "A": [1, 2, 3],
            "B": ["a", "b", "c"],
        },
    )

    for join_type in ["left", "right", "outer"]:
        with pytest.raises(ValueError):
            df.join(df, on="A", strategy="sort_merge", how=join_type)

    with pytest.raises(ValueError):
        df.join(df, on="A", strategy="broadcast", how="outer")


def test_columns_after_join(make_df):
    df1 = make_df(
        {
            "A": [1, 2, 3],
        },
    )

    df2 = make_df({"A": [1, 2, 3], "B": [1, 2, 3]})

    joined_df1 = df1.join(df2, left_on="A", right_on="B")
    joined_df2 = df1.join(df2, left_on="A", right_on="A")

    assert set(joined_df1.schema().column_names()) == set(["A", "B", "right.A"])

    assert set(joined_df2.schema().column_names()) == set(["A", "B"])


def test_rename_join_keys_in_dataframe(make_df):
    df1 = make_df({"A": [1, 2], "B": [2, 2]})

    df2 = make_df({"A": [1, 2]})
    joined_df1 = df1.join(df2, left_on=["A", "B"], right_on=["A", "A"])
    joined_df2 = df1.join(df2, left_on=["B", "A"], right_on=["A", "A"])

    assert set(joined_df1.schema().column_names()) == set(["A", "B"])
    assert set(joined_df2.schema().column_names()) == set(["A", "B"])


@pytest.mark.parametrize("n_partitions", [1, 2, 4])
@pytest.mark.parametrize(
    "join_strategy",
    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
    indirect=True,
)
@pytest.mark.parametrize("join_type", ["inner", "left", "right", "outer"])
def test_joins(join_strategy, join_type, make_df, n_partitions: int, with_morsel_size):
    skip_invalid_join_strategies(join_strategy, join_type)

    df = make_df(
        {
            "A": [1, 2, 3],
            "B": ["a", "b", "c"],
        },
        repartition=n_partitions,
        repartition_columns=["A"],
    )

    joined = df.join(df, on="A", strategy=join_strategy, how=join_type)
    # We shouldn't need to sort the joined output if using a sort-merge join.
    if join_strategy != "sort_merge":
        joined = joined.sort("A")
    joined_data = joined.to_pydict()

    assert joined_data == {
        "A": [1, 2, 3],
        "B": ["a", "b", "c"],
        "right.B": ["a", "b", "c"],
    }


@pytest.mark.parametrize("n_partitions", [1, 2, 4])
@pytest.mark.parametrize(
    "join_strategy",
    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
    indirect=True,
)
@pytest.mark.parametrize("join_type", ["inner", "left", "right", "outer"])
def test_multicol_joins(join_strategy, join_type, make_df, n_partitions: int, with_morsel_size):
    skip_invalid_join_strategies(join_strategy, join_type)

    df = make_df(
        {
            "A": [1, 2, 3],
            "B": ["a", "b", "c"],
            "C": [True, False, True],
        },
        repartition=n_partitions,
        repartition_columns=["A", "B"],
    )

    joined = df.join(df, on=["A", "B"], strategy=join_strategy, how=join_type)
    # We shouldn't need to sort the joined output if using a sort-merge join.
    if join_strategy != "sort_merge":
        joined = joined.sort("A")
    joined_data = joined.to_pydict()

    assert joined_data == {
        "A": [1, 2, 3],
        "B": ["a", "b", "c"],
        "C": [True, False, True],
        "right.C": [True, False, True],
    }


@pytest.mark.parametrize("n_partitions", [1, 2, 4, 8])
@pytest.mark.parametrize(
    "join_strategy",
    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
    indirect=True,
)
@pytest.mark.parametrize("join_type", ["inner", "left", "right", "outer"])
def test_dupes_join_key(join_strategy, join_type, make_df, n_partitions: int, with_morsel_size):
    skip_invalid_join_strategies(join_strategy, join_type)

    df = make_df(
        {
            "A": [1, 1, 2, 2, 3, 3],
            "B": ["a", "b", "c", "d", "e", "f"],
        },
        repartition=n_partitions,
        repartition_columns=["A"],
    )

    joined = df.join(df, on="A", strategy=join_strategy, how=join_type)
    joined = joined.sort(["A", "B", "right.B"])
    joined_data = joined.to_pydict()

    assert joined_data == {
        "A": [1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3],
        "B": ["a", "a", "b", "b", "c", "c", "d", "d", "e", "e", "f", "f"],
        "right.B": ["a", "b", "a", "b", "c", "d", "c", "d", "e", "f", "e", "f"],
    }


@pytest.mark.parametrize("n_partitions", [1, 2, 4, 8])
@pytest.mark.parametrize(
    "join_strategy",
    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
    indirect=True,
)
@pytest.mark.parametrize("join_type", ["inner", "left", "right", "outer"])
def test_multicol_dupes_join_key(join_strategy, join_type, make_df, n_partitions: int, with_morsel_size):
    skip_invalid_join_strategies(join_strategy, join_type)

    df = make_df(
        {
            "A": [1, 1, 2, 2, 3, 3],
            "B": ["a", "a", "b", "b", "c", "d"],
            "C": [1, 0, 1, 0, 1, 0],
        },
        repartition=n_partitions,
        repartition_columns=["A", "B"],
    )

    joined = df.join(df, on=["A", "B"], strategy=join_strategy, how=join_type)
    joined = joined.sort(["A", "B", "C", "right.C"])
    joined_data = joined.to_pydict()

    assert joined_data == {
        "A": [1, 1, 1, 1, 2, 2, 2, 2, 3, 3],
        "B": ["a"] * 4 + ["b"] * 4 + ["c", "d"],
        "C": [0, 0, 1, 1, 0, 0, 1, 1, 1, 0],
        "right.C": [0, 1, 0, 1, 0, 1, 0, 1, 1, 0],
    }


@pytest.mark.parametrize("n_partitions", [1, 2, 4, 6])
@pytest.mark.parametrize(
    "join_strategy",
    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
    indirect=True,
)
@pytest.mark.parametrize("join_type", ["inner", "left", "right", "outer"])
def test_joins_all_same_key(join_strategy, join_type, make_df, n_partitions: int, with_morsel_size):
    skip_invalid_join_strategies(join_strategy, join_type)

    df = make_df(
        {
            "A": [1] * 4,
            "B": ["a", "b", "c", "d"],
        },
        repartition=n_partitions,
        repartition_columns=["A"],
    )

    joined = df.join(df, on="A", strategy=join_strategy, how=join_type)
    joined = joined.sort(["A", "B", "right.B"])
    joined_data = joined.to_pydict()

    assert joined_data == {
        "A": [1] * 16,
        "B": ["a"] * 4 + ["b"] * 4 + ["c"] * 4 + ["d"] * 4,
        "right.B": ["a", "b", "c", "d"] * 4,
    }


@pytest.mark.parametrize("n_partitions", [1, 2, 4])
@pytest.mark.parametrize(
    "join_strategy",
    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
    indirect=True,
)
@pytest.mark.parametrize(
    "join_type,flip,expected",
    [
        ("inner", False, {"A": [], "B": [], "right.B": []}),
        ("inner", True, {"A": [], "B": [], "right.B": []}),
        (
            "left",
            False,
            {"A": [1, 2, 3], "B": ["a", "b", "c"], "right.B": [None, None, None]},
        ),
        (
            "left",
            True,
            {"A": [4, 5, 6], "B": ["d", "e", "f"], "right.B": [None, None, None]},
        ),
        (
            "right",
            False,
            {"A": [4, 5, 6], "B": [None, None, None], "right.B": ["d", "e", "f"]},
        ),
        (
            "right",
            True,
            {"A": [1, 2, 3], "B": [None, None, None], "right.B": ["a", "b", "c"]},
        ),
        (
            "outer",
            False,
            {
                "A": [1, 2, 3, 4, 5, 6],
                "B": ["a", "b", "c", None, None, None],
                "right.B": [None, None, None, "d", "e", "f"],
            },
        ),
        (
            "outer",
            True,
            {
                "A": [1, 2, 3, 4, 5, 6],
                "B": [None, None, None, "d", "e", "f"],
                "right.B": ["a", "b", "c", None, None, None],
            },
        ),
    ],
)
def test_joins_no_overlap_disjoint(
    join_strategy, join_type, flip, expected, make_df, n_partitions: int, with_morsel_size
):
    skip_invalid_join_strategies(join_strategy, join_type)

    df1 = make_df(
        {
            "A": [1, 2, 3],
            "B": ["a", "b", "c"],
        },
        repartition=n_partitions,
        repartition_columns=["A"],
    )
    df2 = make_df(
        {
            "A": [4, 5, 6],
            "B": ["d", "e", "f"],
        },
        repartition=n_partitions,
        repartition_columns=["A"],
    )

    if flip:
        joined = df2.join(df1, on="A", strategy=join_strategy, how=join_type)
    else:
        joined = df1.join(df2, on="A", strategy=join_strategy, how=join_type)
    joined = joined.sort("A")
    joined_data = joined.to_pydict()

    assert joined_data == expected


@pytest.mark.parametrize("n_partitions", [1, 2, 4])
@pytest.mark.parametrize(
    "join_strategy",
    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
    indirect=True,
)
@pytest.mark.parametrize(
    "join_type,flip,expected",
    [
        ("inner", False, {"A": [], "B": [], "right.B": []}),
        ("inner", True, {"A": [], "B": [], "right.B": []}),
        (
            "left",
            False,
            {"A": [1, 3, 5], "B": ["a", "b", "c"], "right.B": [None, None, None]},
        ),
        (
            "left",
            True,
            {"A": [2, 4, 6], "B": ["d", "e", "f"], "right.B": [None, None, None]},
        ),
        (
            "right",
            False,
            {"A": [2, 4, 6], "B": [None, None, None], "right.B": ["d", "e", "f"]},
        ),
        (
            "right",
            True,
            {"A": [1, 3, 5], "B": [None, None, None], "right.B": ["a", "b", "c"]},
        ),
        (
            "outer",
            False,
            {
                "A": [1, 2, 3, 4, 5, 6],
                "B": ["a", None, "b", None, "c", None],
                "right.B": [None, "d", None, "e", None, "f"],
            },
        ),
        (
            "outer",
            True,
            {
                "A": [1, 2, 3, 4, 5, 6],
                "B": [None, "d", None, "e", None, "f"],
                "right.B": ["a", None, "b", None, "c", None],
            },
        ),
    ],
)
def test_joins_no_overlap_interleaved(
    join_strategy, join_type, flip, expected, make_df, n_partitions: int, with_morsel_size
):
    skip_invalid_join_strategies(join_strategy, join_type)

    df1 = make_df(
        {
            "A": [1, 3, 5],
            "B": ["a", "b", "c"],
        },
        repartition=n_partitions,
        repartition_columns=["A"],
    )
    df2 = make_df(
        {
            "A": [2, 4, 6],
            "B": ["d", "e", "f"],
        },
        repartition=n_partitions,
        repartition_columns=["A"],
    )

    if flip:
        joined = df2.join(df1, on="A", strategy=join_strategy, how=join_type)
    else:
        joined = df1.join(df2, on="A", strategy=join_strategy, how=join_type)
    # We shouldn't need to sort the joined output if using a sort-merge join.
    if join_strategy != "sort_merge":
        joined = joined.sort("A")
    joined_data = joined.to_pydict()

    assert joined_data == expected


@pytest.mark.parametrize("n_partitions", [1, 2, 4])
@pytest.mark.parametrize(
    "join_strategy",
    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
    indirect=True,
)
@pytest.mark.parametrize("join_type", ["inner", "left", "right", "outer"])
def test_limit_after_join(join_strategy, join_type, make_df, n_partitions: int, with_morsel_size):
    skip_invalid_join_strategies(join_strategy, join_type)

    data = {
        "A": [1, 2, 3],
    }
    df1 = make_df(
        data,
        repartition=n_partitions,
        repartition_columns=["A"],
    )
    df2 = make_df(
        data,
        repartition=n_partitions,
        repartition_columns=["A"],
    )

    joined = df1.join(df2, on="A", strategy=join_strategy, how=join_type).limit(1)
    joined_data = joined.to_pydict()
    assert "A" in joined_data
    assert len(joined_data["A"]) == 1


###
# Tests for nulls
###


@pytest.mark.parametrize("repartition_nparts", [1, 2, 4])
@pytest.mark.parametrize(
    "join_strategy",
    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
    indirect=True,
)
@pytest.mark.parametrize(
    "join_type,expected",
    [
        (
            "inner",
            {"id": [1, 3], "values_left": ["a1", "c1"], "values_right": ["a2", "c2"]},
        ),
        (
            "left",
            {
                "id": [1, 3, None],
                "values_left": ["a1", "c1", "b1"],
                "values_right": ["a2", "c2", None],
            },
        ),
        (
            "right",
            {
                "id": [1, 2, 3],
                "values_left": ["a1", None, "c1"],
                "values_right": ["a2", "b2", "c2"],
            },
        ),
        (
            "outer",
            {
                "id": [1, 2, 3, None],
                "values_left": ["a1", None, "c1", "b1"],
                "values_right": ["a2", "b2", "c2", None],
            },
        ),
    ],
)
def test_join_with_null(join_strategy, join_type, expected, make_df, repartition_nparts, with_morsel_size):
    skip_invalid_join_strategies(join_strategy, join_type)

    daft_df = make_df(
        {
            "id": [1, None, 3],
            "values_left": ["a1", "b1", "c1"],
        },
        repartition=repartition_nparts,
    )
    daft_df2 = make_df(
        {
            "id": [1, 2, 3],
            "values_right": ["a2", "b2", "c2"],
        },
        repartition=repartition_nparts,
    )
    daft_df = daft_df.join(daft_df2, on="id", strategy=join_strategy, how=join_type)

    assert sort_arrow_table(pa.Table.from_pydict(daft_df.to_pydict()), "id") == sort_arrow_table(
        pa.Table.from_pydict(expected), "id"
    )


@pytest.mark.parametrize("repartition_nparts", [1, 2, 4])
@pytest.mark.parametrize(
    "join_strategy",
    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
    indirect=True,
)
@pytest.mark.parametrize(
    "join_type,expected",
    [
        (
            "inner",
            {"id": [1], "id2": ["foo1"], "values_left": ["a1"], "values_right": ["c2"]},
        ),
        (
            "left",
            {
                "id": [1, None, None],
                "id2": ["foo1", "foo2", None],
                "values_left": ["a1", "b1", "c1"],
                "values_right": ["c2", None, None],
            },
        ),
        (
            "right",
            {
                "id": [1, None, None],
                "id2": ["foo1", "foo2", None],
                "values_left": ["a1", None, None],
                "values_right": ["c2", "a2", "b2"],
            },
        ),
        (
            "outer",
            {
                "id": [1, None, None, None, None],
                "id2": ["foo1", "foo2", "foo2", None, None],
                "values_left": ["a1", "b1", None, "c1", None],
                "values_right": ["c2", None, "a2", None, "b2"],
            },
        ),
    ],
)
def test_join_with_null_multikey(join_strategy, join_type, expected, make_df, repartition_nparts, with_morsel_size):
    skip_invalid_join_strategies(join_strategy, join_type)

    daft_df = make_df(
        {
            "id": [1, None, None],
            "id2": ["foo1", "foo2", None],
            "values_left": ["a1", "b1", "c1"],
        },
        repartition=repartition_nparts,
    )
    daft_df2 = make_df(
        {
            "id": [None, None, 1],
            "id2": ["foo2", None, "foo1"],
            "values_right": ["a2", "b2", "c2"],
        },
        repartition=repartition_nparts,
    )
    daft_df = daft_df.join(daft_df2, on=["id", "id2"], strategy=join_strategy, how=join_type).sort(
        ["id", "id2", "values_left", "values_right"]
    )

    assert sort_arrow_table(pa.Table.from_pydict(daft_df.to_pydict()), "id") == sort_arrow_table(
        pa.Table.from_pydict(expected), "id"
    )


@pytest.mark.parametrize("repartition_nparts", [1, 2, 4])
@pytest.mark.parametrize(
    "join_strategy",
    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
    indirect=True,
)
@pytest.mark.parametrize(
    "join_type,expected",
    [
        (
            "inner",
            {
                "left_id": [1],
                "left_id2": ["foo1"],
                "values_left": ["a1"],
                "right_id": [1],
                "right_id2": ["foo1"],
                "values_right": ["c2"],
            },
        ),
        (
            "left",
            {
                "left_id": [1, None, None],
                "left_id2": ["foo1", "foo2", None],
                "values_left": ["a1", "b1", "c1"],
                "right_id": [1, None, None],
                "right_id2": ["foo1", None, None],
                "values_right": ["c2", None, None],
            },
        ),
        (
            "right",
            {
                "left_id": [1, None, None],
                "left_id2": ["foo1", None, None],
                "values_left": ["a1", None, None],
                "right_id": [1, None, None],
                "right_id2": ["foo1", "foo2", None],
                "values_right": ["c2", "a2", "b2"],
            },
        ),
        (
            "outer",
            {
                "left_id": [1, None, None, None, None],
                "left_id2": ["foo1", "foo2", None, None, None],
                "values_left": ["a1", "b1", None, "c1", None],
                "right_id": [1, None, None, None, None],
                "right_id2": ["foo1", None, "foo2", None, None],
                "values_right": ["c2", None, "a2", None, "b2"],
            },
        ),
    ],
)
def test_join_with_null_asymmetric_multikey(
    join_strategy, join_type, expected, make_df, repartition_nparts, with_morsel_size
):
    skip_invalid_join_strategies(join_strategy, join_type)

    daft_df = make_df(
        {
            "left_id": [1, None, None],
            "left_id2": ["foo1", "foo2", None],
            "values_left": ["a1", "b1", "c1"],
        },
        repartition=repartition_nparts,
    )
    daft_df2 = make_df(
        {
            "right_id": [None, None, 1],
            "right_id2": ["foo2", None, "foo1"],
            "values_right": ["a2", "b2", "c2"],
        },
        repartition=repartition_nparts,
    )
    daft_df = daft_df.join(
        daft_df2,
        left_on=["left_id", "left_id2"],
        right_on=["right_id", "right_id2"],
        how=join_type,
        strategy=join_strategy,
    ).sort(["left_id", "left_id2", "right_id", "right_id2", "values_left", "values_right"])

    assert sort_arrow_table(pa.Table.from_pydict(daft_df.to_pydict()), "left_id") == sort_arrow_table(
        pa.Table.from_pydict(expected), "left_id"
    )


@pytest.mark.parametrize("repartition_nparts", [1, 2, 4])
@pytest.mark.parametrize(
    "join_strategy",
    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
    indirect=True,
)
@pytest.mark.parametrize(
    "join_type,expected",
    [
        (
            "inner",
            {
                "id": [],
                "values_left": [],
                "values_right": [],
            },
        ),
        (
            "left",
            {
                "id": [None, None, None],
                "values_left": ["a1", "b1", "c1"],
                "values_right": [None, None, None],
            },
        ),
        (
            "right",
            {
                "id": [1, 2, 3],
                "values_left": [None, None, None],
                "values_right": ["a2", "b2", "c2"],
            },
        ),
        (
            "outer",
            {
                "id": [1, 2, 3, None, None, None],
                "values_left": [None, None, None, "a1", "b1", "c1"],
                "values_right": ["a2", "b2", "c2", None, None, None],
            },
        ),
    ],
)
def test_join_all_null(join_strategy, join_type, expected, make_df, repartition_nparts, with_morsel_size):
    skip_invalid_join_strategies(join_strategy, join_type)

    daft_df = make_df(
        {
            "id": [None, None, None],
            "values_left": ["a1", "b1", "c1"],
        },
        repartition=repartition_nparts,
    )
    daft_df2 = make_df(
        {
            "id": [1, 2, 3],
            "values_right": ["a2", "b2", "c2"],
        },
        repartition=repartition_nparts,
    )
    daft_df = (
        daft_df.with_column("id", daft_df["id"].cast(DataType.int64()))
        .join(daft_df2, on="id", how=join_type, strategy=join_strategy)
        .sort(["id", "values_left", "values_right"])
    )

    assert sort_arrow_table(pa.Table.from_pydict(daft_df.to_pydict()), "id") == sort_arrow_table(
        pa.Table.from_pydict(expected), "id"
    )


@pytest.mark.parametrize(
    "join_strategy",
    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
    indirect=True,
)
@pytest.mark.parametrize(
    "join_type,expected",
    [
        ("inner", {"id": [], "values_left": [], "values_right": []}),
        ("left", {"id": [None, None, None], "values_left": ["a1", "b1", "c1"], "values_right": [None, None, None]}),
        ("right", {"id": [None, None, None], "values_left": [None, None, None], "values_right": ["a2", "b2", "c2"]}),
        (
            "outer",
            {
                "id": [None, None, None, None, None, None],
                "values_left": ["a1", "b1", "c1", None, None, None],
                "values_right": [None, None, None, "a2", "b2", "c2"],
            },
        ),
    ],
)
def test_join_null_type_column(join_strategy, join_type, expected, make_df, with_morsel_size):
    skip_invalid_join_strategies(join_strategy, join_type)

    daft_df = make_df(
        {
            "id": [None, None, None],
            "values_left": ["a1", "b1", "c1"],
        }
    )
    daft_df2 = make_df(
        {
            "id": [None, None, None],
            "values_right": ["a2", "b2", "c2"],
        }
    )

    daft_df = daft_df.join(daft_df2, on="id", how=join_type, strategy=join_strategy).sort(
        ["values_left", "values_right"]
    )
    assert pa.Table.from_pydict(daft_df.to_pydict()) == pa.Table.from_pydict(expected)


@pytest.mark.parametrize("repartition_nparts", [1, 2, 4])
@pytest.mark.parametrize(
    "join_strategy",
    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
    indirect=True,
)
@pytest.mark.parametrize(
    "join_type,expected",
    [
        (
            "semi",
            {
                "id": [2, 3],
                "values_left": ["b1", "c1"],
            },
        ),
        (
            "anti",
            {
                "id": [1, None],
                "values_left": ["a1", "d1"],
            },
        ),
    ],
)
def test_join_semi_anti(join_strategy, join_type, expected, make_df, repartition_nparts, with_morsel_size):
    skip_invalid_join_strategies(join_strategy, join_type)

    daft_df1 = make_df(
        {
            "id": [1, 2, 3, None],
            "values_left": ["a1", "b1", "c1", "d1"],
        },
        repartition=repartition_nparts,
    )
    daft_df2 = make_df(
        {
            "id": [2, 2, 3, 4],
            "values_right": ["a2", "b2", "c2", "d2"],
        },
        repartition=repartition_nparts,
    )
    daft_df = (
        daft_df1.with_column("id", daft_df1["id"].cast(DataType.int64()))
        .join(daft_df2, on="id", how=join_type, strategy=join_strategy)
        .sort(["id", "values_left"])
    ).select("id", "values_left")

    assert sort_arrow_table(pa.Table.from_pydict(daft_df.to_pydict()), "id") == sort_arrow_table(
        pa.Table.from_pydict(expected), "id"
    )


@pytest.mark.parametrize("repartition_nparts", [1, 2, 4])
@pytest.mark.parametrize(
    "join_strategy",
    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
    indirect=True,
)
@pytest.mark.parametrize(
    "join_type,expected",
    [
        (
            "semi",
            {
                "id_left": [2, 3],
                "values_left": ["b1", "c1"],
            },
        ),
        (
            "anti",
            {
                "id_left": [1, None],
                "values_left": ["a1", "d1"],
            },
        ),
    ],
)
def test_join_semi_anti_different_names(
    join_strategy, join_type, expected, make_df, repartition_nparts, with_morsel_size
):
    skip_invalid_join_strategies(join_strategy, join_type)

    daft_df1 = make_df(
        {
            "id_left": [1, 2, 3, None],
            "values_left": ["a1", "b1", "c1", "d1"],
        },
        repartition=repartition_nparts,
    )
    daft_df2 = make_df(
        {
            "id_right": [2, 2, 3, 4],
            "values_right": ["a2", "b2", "c2", "d2"],
        },
        repartition=repartition_nparts,
    )
    daft_df = (
        daft_df1.with_column("id_left", daft_df1["id_left"].cast(DataType.int64()))
        .join(
            daft_df2,
            left_on="id_left",
            right_on="id_right",
            how=join_type,
            strategy=join_strategy,
        )
        .sort(["id_left", "values_left"])
    ).select("id_left", "values_left")

    assert sort_arrow_table(pa.Table.from_pydict(daft_df.to_pydict()), "id_left") == sort_arrow_table(
        pa.Table.from_pydict(expected), "id_left"
    )


@pytest.mark.parametrize(
    "join_type,expected_dtypes",
    [
        (
            "inner",
            {
                "id": DataType.int64(),
                "values": DataType.string(),
                "right.values": DataType.string(),
            },
        ),
        (
            "left",
            {
                "id": DataType.int64(),
                "values": DataType.string(),
                "right.values": DataType.string(),
            },
        ),
        (
            "right",
            {
                "id": DataType.float64(),
                "values": DataType.string(),
                "right.values": DataType.string(),
            },
        ),
        (
            "outer",
            {
                "id": DataType.float64(),
                "values": DataType.string(),
                "right.values": DataType.string(),
            },
        ),
    ],
)
def test_join_true_join_keys(join_type, expected_dtypes, make_df, with_morsel_size):
    daft_df = make_df(
        {
            "id": [1, 2, 3],
            "values": ["a", "b", "c"],
        }
    )
    daft_df2 = make_df(
        {
            "id": [2.0, 2.5, 3.0, 4.0],
            "values": ["a2", "b2", "c2", "d2"],
        }
    )

    result = daft_df.join(daft_df2, left_on=["id", "values"], right_on=["id", col("values").str.left(1)], how=join_type)

    assert result.schema().column_names() == ["id", "values", "right.values"]
    assert result.schema()["id"].dtype == expected_dtypes["id"]
    assert result.schema()["values"].dtype == expected_dtypes["values"]
    assert result.schema()["right.values"].dtype == expected_dtypes["right.values"]


@pytest.mark.parametrize(
    "join_strategy",
    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
    indirect=True,
)
@pytest.mark.parametrize(
    "join_type,expected",
    [
        (
            "inner",
            {
                "a": [2, 3],
                "b": [2, 3],
            },
        ),
        (
            "left",
            {
                "a": [1, 2, 3],
                "b": [None, 2, 3],
            },
        ),
        (
            "right",
            {
                "a": [2, 3, None],
                "b": [2, 3, 4],
            },
        ),
        (
            "outer",
            {
                "a": [1, 2, 3, None],
                "b": [None, 2, 3, 4],
            },
        ),
        (
            "semi",
            {
                "a": [2, 3],
            },
        ),
        (
            "anti",
            {
                "a": [1],
            },
        ),
    ],
)
def test_join_with_alias_in_key(join_strategy, join_type, expected, make_df, with_morsel_size):
    skip_invalid_join_strategies(join_strategy, join_type)

    daft_df1 = make_df(
        {
            "a": [1, 2, 3],
        }
    )
    daft_df2 = make_df(
        {
            "b": [2, 3, 4],
        }
    )

    daft_df = daft_df1.join(daft_df2, left_on=col("a").alias("x"), right_on="b", how=join_type, strategy=join_strategy)

    assert sort_arrow_table(pa.Table.from_pydict(daft_df.to_pydict()), "a") == sort_arrow_table(
        pa.Table.from_pydict(expected), "a"
    )


@pytest.mark.parametrize(
    "join_strategy",
    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
    indirect=True,
)
@pytest.mark.parametrize(
    "join_type,expected",
    [
        (
            "inner",
            {
                "a": [2, 3],
                "right.a": [2, 3],
            },
        ),
        (
            "left",
            {
                "a": [1, 2, 3],
                "right.a": [None, 2, 3],
            },
        ),
        (
            "right",
            {
                "a": [2, 3, None],
                "right.a": [2, 3, 4],
            },
        ),
        (
            "outer",
            {
                "a": [1, 2, 3, None],
                "right.a": [None, 2, 3, 4],
            },
        ),
        (
            "semi",
            {
                "a": [2, 3],
            },
        ),
        (
            "anti",
            {
                "a": [1],
            },
        ),
    ],
)
def test_join_same_name_alias(join_strategy, join_type, expected, make_df, with_morsel_size):
    skip_invalid_join_strategies(join_strategy, join_type)

    daft_df1 = make_df(
        {
            "a": [1, 2, 3],
        }
    )
    daft_df2 = make_df(
        {
            "a": [2, 3, 4],
        }
    )

    daft_df = daft_df1.join(daft_df2, left_on="a", right_on=col("a").alias("b"), how=join_type, strategy=join_strategy)

    assert sort_arrow_table(pa.Table.from_pydict(daft_df.to_pydict()), "a") == sort_arrow_table(
        pa.Table.from_pydict(expected), "a"
    )


@pytest.mark.parametrize(
    "join_strategy",
    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
    indirect=True,
)
@pytest.mark.parametrize(
    "join_type,expected",
    [
        (
            "inner",
            {
                "a": [0.2, 0.3],
                "right.a": [20, 30],
            },
        ),
        (
            "left",
            {
                "a": [0.1, 0.2, 0.3],
                "right.a": [None, 20, 30],
            },
        ),
        (
            "right",
            {
                "a": [0.2, 0.3, None],
                "right.a": [20, 30, 40],
            },
        ),
        (
            "outer",
            {
                "a": [0.1, 0.2, 0.3, None],
                "right.a": [None, 20, 30, 40],
            },
        ),
        (
            "semi",
            {
                "a": [0.2, 0.3],
            },
        ),
        (
            "anti",
            {
                "a": [0.1],
            },
        ),
    ],
)
def test_join_same_name_alias_with_compute(join_strategy, join_type, expected, make_df, with_morsel_size):
    skip_invalid_join_strategies(join_strategy, join_type)

    daft_df1 = make_df(
        {
            "a": [0.1, 0.2, 0.3],
        }
    )
    daft_df2 = make_df(
        {
            "a": [20, 30, 40],
        }
    )

    daft_df = daft_df1.join(
        daft_df2, left_on=col("a") * 10, right_on=(col("a") / 10).alias("b"), how=join_type, strategy=join_strategy
    )

    assert sort_arrow_table(pa.Table.from_pydict(daft_df.to_pydict()), "a") == sort_arrow_table(
        pa.Table.from_pydict(expected), "a"
    )


@pytest.mark.parametrize(
    "suffix,prefix,expected",
    [
        (None, None, "right.score"),
        ("_right", None, "score_right"),
        (None, "left_", "left_score"),
        ("_right", "prefix.", "prefix.score_right"),
    ],
)
def test_join_suffix_and_prefix(suffix, prefix, expected, make_df, with_morsel_size):
    df1 = daft.from_pydict({"idx": [1, 2], "val": [10, 20]})
    df2 = daft.from_pydict({"idx": [3], "score": [0.1]})
    df3 = daft.from_pydict({"idx": [1], "score": [0.1]})

    df = df1.join(df2, on="idx").join(df3, on="idx", suffix=suffix, prefix=prefix)
    assert df.column_names == ["idx", "val", "score", expected]


@pytest.mark.parametrize("left_partitions", [1, 2, 4])
@pytest.mark.parametrize("right_partitions", [1, 2, 4])
def test_cross_join(left_partitions, right_partitions, make_df, with_morsel_size):
    df1 = make_df(
        {
            "A": [1, 3, 5],
            "B": ["a", "b", "c"],
        },
        repartition=left_partitions,
        repartition_columns=["A"],
    )

    df2 = make_df(
        {
            "C": [2, 4, 6, 8],
            "D": ["d", "e", "f", "g"],
        },
        repartition=right_partitions,
        repartition_columns=["C"],
    )

    df = df1.join(df2, how="cross")
    df = df.sort(["A", "C"])

    assert df.to_pydict() == {
        "A": [1, 1, 1, 1, 3, 3, 3, 3, 5, 5, 5, 5],
        "B": ["a", "a", "a", "a", "b", "b", "b", "b", "c", "c", "c", "c"],
        "C": [2, 4, 6, 8, 2, 4, 6, 8, 2, 4, 6, 8],
        "D": ["d", "e", "f", "g", "d", "e", "f", "g", "d", "e", "f", "g"],
    }


@pytest.mark.parametrize("join_type", ["inner", "left", "right", "outer", "anti", "semi", "cross"])
@pytest.mark.parametrize("repartition_nparts", [1, 2, 4])
@pytest.mark.parametrize(
    "left,right,expected",
    [
        # Case 1: Left has data, right is empty
        (
            {"a": [1, 2, 3, 4], "b": ["a", "b", "c", "d"]},
            {"c": [], "d": []},
            {
                "inner": {"a": [], "b": [], "c": [], "d": []},
                "left": {
                    "a": [1, 2, 3, 4],
                    "b": ["a", "b", "c", "d"],
                    "c": [None, None, None, None],
                    "d": [None, None, None, None],
                },
                "right": {"a": [], "b": [], "c": [], "d": []},
                "outer": {
                    "a": [1, 2, 3, 4],
                    "b": ["a", "b", "c", "d"],
                    "c": [None, None, None, None],
                    "d": [None, None, None, None],
                },
                "anti": {"a": [1, 2, 3, 4], "b": ["a", "b", "c", "d"]},
                "semi": {"a": [], "b": []},
                "cross": {"a": [], "b": [], "c": [], "d": []},
            },
        ),
        # Case 2: Left is empty, right has data
        (
            {"a": [], "b": []},
            {"c": [5, 6, 7], "d": ["e", "f", "g"]},
            {
                "inner": {"a": [], "b": [], "c": [], "d": []},
                "left": {"a": [], "b": [], "c": [], "d": []},
                "right": {"a": [None, None, None], "b": [None, None, None], "c": [5, 6, 7], "d": ["e", "f", "g"]},
                "outer": {"a": [None, None, None], "b": [None, None, None], "c": [5, 6, 7], "d": ["e", "f", "g"]},
                "anti": {"a": [], "b": []},
                "semi": {"a": [], "b": []},
                "cross": {"a": [], "b": [], "c": [], "d": []},
            },
        ),
        # Case 3: Both empty
        (
            {"a": [], "b": []},
            {"c": [], "d": []},
            {
                "inner": {"a": [], "b": [], "c": [], "d": []},
                "left": {"a": [], "b": [], "c": [], "d": []},
                "right": {"a": [], "b": [], "c": [], "d": []},
                "outer": {"a": [], "b": [], "c": [], "d": []},
                "anti": {"a": [], "b": []},
                "semi": {"a": [], "b": []},
                "cross": {"a": [], "b": [], "c": [], "d": []},
            },
        ),
    ],
)
def test_join_empty(join_type, repartition_nparts, left, right, expected, make_df, with_morsel_size):
    left_df = make_df(
        left,
        repartition=repartition_nparts,
        repartition_columns=["a"],
    )

    right_df = make_df(
        right,
        repartition=repartition_nparts,
        repartition_columns=["c"],
    )

    if join_type == "cross":
        left_on = None
        right_on = None
    else:
        left_on = ["a"]
        right_on = ["c"]

    result = left_df.join(right_df, left_on=left_on, right_on=right_on, how=join_type)

    sort_by = ["a", "b", "c", "d"] if join_type in ["inner", "left", "right", "outer", "cross"] else ["a", "b"]

    assert sort_arrow_table(pa.Table.from_pydict(result.to_pydict()), *sort_by) == sort_arrow_table(
        pa.Table.from_pydict(expected[join_type]), *sort_by
    )


@pytest.mark.parametrize(
    "join_type,expected",
    [
        ("inner", {"a": [1, 1, 2], "b": ["a", "a", "b"], "c": [1.0, 1.0, 2.0], "d": ["g", "h", "j"]}),
        (
            "left",
            {
                "a": [1, 1, 2, 3, 4, 5, 6],
                "b": ["a", "a", "b", "c", "d", "e", "f"],
                "c": [1.0, 1.0, 2.0, None, None, None, None],
                "d": ["g", "h", "j", None, None, None, None],
            },
        ),
        (
            "right",
            {
                "a": [1, 1, 2, None, None, None],
                "b": ["a", "a", "b", None, None, None],
                "c": [1.0, 1.0, 2.0, 1.5, 2.1, 2.5],
                "d": ["g", "h", "j", "i", "k", "l"],
            },
        ),
        (
            "outer",
            {
                "a": [1, 1, 2, 3, 4, 5, 6, None, None, None],
                "b": ["a", "a", "b", "c", "d", "e", "f", None, None, None],
                "c": [1.0, 1.0, 2.0, None, None, None, None, 1.5, 2.1, 2.5],
                "d": ["g", "h", "j", None, None, None, None, "i", "k", "l"],
            },
        ),
        (
            "anti",
            {
                "a": [3, 4, 5, 6],
                "b": ["c", "d", "e", "f"],
            },
        ),
        (
            "semi",
            {
                "a": [1, 2],
                "b": ["a", "b"],
            },
        ),
    ],
)
@pytest.mark.parametrize("repartition_nparts", [1, 2, 4])
def test_join_different_join_key_types(join_type, expected, repartition_nparts, make_df, with_morsel_size):
    left_df = make_df(
        {"a": [1, 2, 3, 4, 5, 6], "b": ["a", "b", "c", "d", "e", "f"]},
        repartition=repartition_nparts,
        repartition_columns=["a"],
    )

    right_df = make_df(
        {"c": [1.0, 1.0, 1.5, 2.0, 2.1, 2.5], "d": ["g", "h", "i", "j", "k", "l"]},
        repartition=repartition_nparts,
        repartition_columns=["c"],
    )

    result = left_df.join(right_df, left_on="a", right_on="c", how=join_type)

    sort_by = ["a", "b", "c", "d"] if join_type in ["inner", "left", "right", "outer"] else ["a", "b"]

    assert sort_arrow_table(pa.Table.from_pydict(result.to_pydict()), *sort_by) == sort_arrow_table(
        pa.Table.from_pydict(expected), *sort_by
    )


# Tests issue 4113.
def test_self_join_with_projection():
    df1 = daft.from_pydict(
        {
            "pattern_id_l": ["a"],
            "pattern_id_r": ["b"],
        }
    )

    df2 = daft.from_pydict(
        {
            "id": ["1", "2", "3"],
            "pattern_id": ["a", "b", "c"],
        }
    )

    result_df = (
        df1.join(
            df2,
            left_on="pattern_id_l",
            right_on="pattern_id",
        )
        .select(daft.col("id").alias("id_l"), daft.col("pattern_id_r"))
        .join(
            df2,
            left_on="pattern_id_r",
            right_on="pattern_id",
        )
        .select(
            daft.col("id_l"),
            daft.col("id").alias("id_r"),
        )
    )

    expected = {
        "id_l": ["1"],
        "id_r": ["2"],
    }

    assert pa.Table.from_pydict(result_df.to_pydict()) == pa.Table.from_pydict(expected)



from __future__ import annotations

import pytest

from daft.io.partitioning import PartitionField, PartitionTransform
from daft.schema import DataType, Field


def test_partition_field():
    field = Field.create("test_field", DataType.int64())
    source_field = Field.create("source_field", DataType.int64())
    transform = PartitionTransform.identity()

    partition_field = PartitionField.create(field, source_field, transform)
    assert partition_field.field == field
    assert partition_field.source_field == source_field
    assert partition_field.transform == transform

    partition_field = PartitionField.create(field)
    assert partition_field.field == field
    assert partition_field.source_field is None
    assert partition_field.transform is None


@pytest.mark.parametrize(
    "transform_factory,transform_type,transform_args",
    [
        (PartitionTransform.identity, "identity", None),
        (PartitionTransform.year, "year", None),
        (PartitionTransform.month, "month", None),
        (PartitionTransform.day, "day", None),
        (PartitionTransform.hour, "hour", None),
        (PartitionTransform.iceberg_bucket, "iceberg_bucket", [10]),
        (PartitionTransform.iceberg_truncate, "iceberg_truncate", [10]),
    ],
)
def test_partition_transform_types(transform_factory, transform_type, transform_args):
    if transform_args:
        transform = transform_factory(*transform_args)
    else:
        transform = transform_factory()

    for other_transform_type in ["identity", "year", "month", "day", "hour", "iceberg_bucket", "iceberg_truncate"]:
        is_other_transform = transform_type == other_transform_type
        assert getattr(transform, f"is_{other_transform_type}")() == is_other_transform

    if transform_type == "iceberg_bucket":
        assert transform.num_buckets == transform_args[0]

    if transform_type == "iceberg_truncate":
        assert transform.width == transform_args[0]


def test_partition_transform_direct_initialization():
    with pytest.raises(ValueError, match="We do not support creating a PartitionTransform directly"):
        PartitionTransform()


def test_partition_transform_repr():
    transform = PartitionTransform.identity()
    assert repr(transform) == transform._partition_transform.__repr__()



from __future__ import annotations

import urllib.parse
from typing import TYPE_CHECKING

from daft import Series

from .micropartition import MicroPartition

if TYPE_CHECKING:
    from daft.expressions import ExpressionsProjection
    from daft.recordbatch.recordbatch import RecordBatch


def partition_strings_to_path(
    root_path: str,
    parts: dict[str, str],
    partition_null_fallback: str = "__HIVE_DEFAULT_PARTITION__",
) -> str:
    # Hive-style partition keys and values should be URL encoded.
    keys = [urllib.parse.quote(key) for key in parts.keys()]
    values = [partition_null_fallback if value is None else urllib.parse.quote(value) for value in parts.values()]
    postfix = "/".join(f"{k}={v}" for k, v in zip(keys, values))
    return f"{root_path}/{postfix}"


def partition_values_to_str_mapping(
    partition_values: MicroPartition | RecordBatch,
) -> dict[str, Series]:
    null_part = Series.from_pylist(
        [None]
    )  # This is to ensure that the null values are replaced with the default_partition_fallback value

    partition_strings = {}

    for column in partition_values.columns():
        string_names = column._to_str_values()
        null_filled = column.is_null().if_else(null_part, string_names)
        partition_strings[column.name()] = null_filled

    return partition_strings


class PartitionedTable:
    def __init__(self, table: MicroPartition, partition_keys: ExpressionsProjection | None):
        self.table: MicroPartition = table
        self.partition_keys: ExpressionsProjection | None = partition_keys
        self._partitions: list[MicroPartition] | None = None
        self._partition_values: MicroPartition | None = None

    def _create_partitions(self) -> None:
        if self.partition_keys is None or len(self.partition_keys) == 0:
            self._partitions = [self.table]
            self._partition_values = None
        else:
            self._partitions, self._partition_values = self.table.partition_by_value(partition_keys=self.partition_keys)

    def partitions(self) -> list[MicroPartition]:
        """Returns a list of MicroPartitions representing the table partitioned by the partition keys.

        If the table is not partitioned, returns the original table as the single element in the list.
        """
        if self._partitions is None:
            self._create_partitions()
        return self._partitions  # type: ignore

    def partition_values(self) -> MicroPartition | None:
        """Returns the partition values, with each row corresponding to the partition at the same index in PartitionedTable.partitions().

        If the table is not partitioned, returns None.

        """
        if self._partition_values is None:
            self._create_partitions()
        return self._partition_values

    def partition_values_str(self) -> MicroPartition | None:
        """Returns the partition values converted to human-readable strings, keeping null values as null.

        If the table is not partitioned, returns None.
        """
        partition_values = self.partition_values()

        if partition_values is None:
            return None
        else:
            partition_strings = partition_values_to_str_mapping(partition_values)
            return MicroPartition.from_pydict(partition_strings)



from __future__ import annotations

import daft
from daft import col


def test_getitem_list():
    df = daft.from_pydict({"a": [[1, 2, 3], [1], [1, 2]]})
    df = df.select(df["a"][0].alias("first"), col("a")[1].alias("second"))
    assert df.to_pydict() == {"first": [1, 1, 1], "second": [2, None, 2]}


def test_getitem_struct():
    df = daft.from_pydict({"a": [{"x": 1, "y": "foo"}, {"x": 2, "y": "bar"}]})
    df = df.select(col("a")["x"], df["a"]["y"])
    assert df.to_pydict() == {"x": [1, 2], "y": ["foo", "bar"]}


def test_getitem_struct_all():
    df = daft.from_pydict({"a": [{"x": 1, "y": "foo"}, {"x": 2, "y": "bar"}]})
    df = df.select(df["a"]["*"])
    assert df.to_pydict() == {"x": [1, 2], "y": ["foo", "bar"]}



from __future__ import annotations

import pytest

import daft


def test_upload_local(tmpdir):
    bytes_data = [b"a", b"b", b"c"]
    data = {"data": bytes_data}
    df = daft.from_pydict(data)
    df = df.with_column("files", df["data"].url.upload(str(tmpdir + "/nested")))
    df.collect()

    results = df.to_pydict()
    assert results["data"] == bytes_data
    assert len(results["files"]) == len(bytes_data)
    for path, expected in zip(results["files"], bytes_data):
        assert path.startswith("file://")
        path = path[len("file://") :]
        with open(path, "rb") as f:
            assert f.read() == expected


def test_upload_local_single_file_url(tmpdir):
    bytes_data = [b"a"]
    paths = [f"{tmpdir}/0"]
    data = {"data": bytes_data, "paths": paths}
    df = daft.from_pydict(data)
    # Even though there is only one row, since we pass in the upload URL via an expression, we
    # should treat the given path as a per-row path and write directly to that path, instead of
    # treating the path as a directory and writing to `{path}/uuid`.
    df = df.with_column("files", df["data"].url.upload(df["paths"]))
    df.collect()

    results = df.to_pydict()
    assert results["data"] == bytes_data
    assert len(results["files"]) == len(bytes_data)
    for path, expected in zip(results["files"], bytes_data):
        assert path.startswith("file://")
        path = path[len("file://") :]
        with open(path, "rb") as f:
            assert f.read() == expected
    # Check that data was uploaded to the correct paths.
    for path, expected in zip(results["files"], paths):
        assert path == "file://" + expected


def test_upload_local_row_specifc_urls(tmpdir):
    bytes_data = [b"a", b"b", b"c"]
    paths = [f"{tmpdir}/0", f"{tmpdir}/1", f"{tmpdir}/2"]
    data = {"data": bytes_data, "paths": paths}
    df = daft.from_pydict(data)
    df = df.with_column("files", df["data"].url.upload(df["paths"]))
    df.collect()

    results = df.to_pydict()
    assert results["data"] == bytes_data
    assert len(results["files"]) == len(bytes_data)
    for path, expected in zip(results["files"], bytes_data):
        assert path.startswith("file://")
        path = path[len("file://") :]
        with open(path, "rb") as f:
            assert f.read() == expected
    # Check that data was uploaded to the correct paths.
    for path, expected in zip(results["files"], paths):
        assert path == "file://" + expected


def test_upload_local_no_write_permissions(tmpdir):
    bytes_data = [b"a", b"b", b"c"]
    # We have no write permissions to the first and third paths.
    paths = ["/some-root-path", f"{tmpdir}/normal_path", "/another-bad-path"]
    expected_paths = [None, f"file://{tmpdir}/normal_path", None]
    expected_data = b"b"
    data = {"data": bytes_data, "paths": paths}
    df = daft.from_pydict(data)
    df_raise_error = df.with_column("files", df["data"].url.upload(df["paths"]))
    with pytest.raises(daft.exceptions.DaftCoreException):
        df_raise_error.collect()
    # Retry with `on_error` set to `null`.
    df_null = df.with_column("files", df["data"].url.upload(df["paths"], on_error="null"))
    df_null.collect()
    results = df_null.to_pydict()
    for path, expected_path in zip(results["files"], expected_paths):
        assert (path is None and expected_path is None) or path == expected_path
        if path is not None:
            assert path.startswith("file://")
            path = path[len("file://") :]
            with open(path, "rb") as f:
                assert f.read() == expected_data



from __future__ import annotations

import os
import socket
import urllib
from unittest.mock import MagicMock, patch

import pytest

from daft.scarf_telemetry import track_import_on_scarf, track_runner_on_scarf


@pytest.fixture
def ensure_analytics_enabled():
    """Ensure DAFT_ANALYTICS_ENABLED is not set to 0 for tests."""
    # Store original value
    original_value = os.environ.get("DAFT_ANALYTICS_ENABLED")

    # Set to a non-zero value
    if original_value == "0":
        os.environ.pop("DAFT_ANALYTICS_ENABLED", None)

    yield

    # Restore original value after test
    if original_value is not None:
        os.environ["DAFT_ANALYTICS_ENABLED"] = original_value
    else:
        os.environ.pop("DAFT_ANALYTICS_ENABLED", None)


@pytest.mark.parametrize(
    "telemetry_fn,endpoint,extra_params",
    [
        (track_runner_on_scarf, "daft-runner", {"runner": "ray"}),
        (track_import_on_scarf, "daft-import", None),
    ],
)
@patch("daft.get_build_type")
@patch("daft.get_version")
@patch("urllib.request.urlopen")
def test_scarf_telemetry_basic(
    mock_urlopen: MagicMock,
    mock_version: MagicMock,
    mock_build_type: MagicMock,
    telemetry_fn,
    endpoint,
    extra_params,
    ensure_analytics_enabled,
):
    # Test basic functionality of scarf_telemetry verify that analytics are successfully sent and url is properly formatted with all required parameters

    # Set up mocks for version and build_type
    mock_version.return_value = "0.0.0"
    mock_build_type.return_value = "release"
    mock_response = MagicMock()
    mock_response.status = 200
    mock_urlopen.return_value.__enter__.return_value = mock_response

    # Test basic analytics call
    if extra_params:
        request_thread, result_container = telemetry_fn(**extra_params)
    else:
        request_thread, result_container = telemetry_fn()

    request_thread.join()

    assert result_container["response_status"] == "Response status: 200"
    if extra_params and "runner" in extra_params:
        assert result_container["extra_value"] == extra_params["runner"]
    else:
        assert result_container["extra_value"] is None

    # Verify URL format and parameters
    called_url = mock_urlopen.call_args[0][0]
    assert called_url.startswith(f"https://daft.gateway.scarf.sh/{endpoint}?")
    assert "version=0.0.0" in called_url
    if extra_params and "runner" in extra_params:
        assert f"runner={extra_params['runner']}" in called_url


@pytest.mark.parametrize(
    "telemetry_fn,extra_params",
    [
        (track_runner_on_scarf, {"runner": "ray"}),
        (track_import_on_scarf, None),
    ],
)
@patch("daft.get_build_type")
@patch("daft.get_version")
def test_scarf_telemetry_dev_build(
    mock_version: MagicMock,
    mock_build_type: MagicMock,
    telemetry_fn,
    extra_params,
    ensure_analytics_enabled,
):
    # Test that analytics are not sent for dev builds, function returns None for both status and runner type

    mock_version.return_value = "0.0.0"
    mock_build_type.return_value = "dev"

    if extra_params:
        request_thread, result_container = telemetry_fn(**extra_params)
    else:
        request_thread, result_container = telemetry_fn()

    assert request_thread is None
    assert result_container["response_status"] is None
    assert result_container["extra_value"] is None


@pytest.mark.parametrize(
    "telemetry_fn,extra_params",
    [
        (track_runner_on_scarf, {"runner": "ray"}),
        (track_import_on_scarf, None),
    ],
)
@patch("daft.get_build_type")
@patch("daft.get_version")
@patch("urllib.request.urlopen")
def test_scarf_telemetry_opt_out_with_scarf_analytics(
    mock_urlopen: MagicMock,
    mock_version: MagicMock,
    mock_build_type: MagicMock,
    telemetry_fn,
    extra_params,
    ensure_analytics_enabled,
):
    mock_version.return_value = "0.0.0"
    mock_build_type.return_value = "release"

    # Clean and set environment
    os.environ.pop("DO_NOT_TRACK", None)
    os.environ["SCARF_NO_ANALYTICS"] = "true"

    if extra_params:
        request_thread, result_container = telemetry_fn(**extra_params)
    else:
        request_thread, result_container = telemetry_fn()

    assert request_thread is None
    assert result_container["response_status"] is None
    assert result_container["extra_value"] is None
    mock_urlopen.assert_not_called()


@pytest.mark.parametrize(
    "telemetry_fn,extra_params",
    [
        (track_runner_on_scarf, {"runner": "ray"}),
        (track_import_on_scarf, None),
    ],
)
@patch("daft.get_build_type")
@patch("daft.get_version")
@patch("urllib.request.urlopen")
def test_scarf_telemetry_opt_out_with_do_not_track(
    mock_urlopen: MagicMock,
    mock_version: MagicMock,
    mock_build_type: MagicMock,
    telemetry_fn,
    extra_params,
    ensure_analytics_enabled,
):
    mock_version.return_value = "0.0.0"
    mock_build_type.return_value = "release"

    # Clean and set environment
    os.environ.pop("SCARF_NO_ANALYTICS", None)
    os.environ["DO_NOT_TRACK"] = "true"

    if extra_params:
        request_thread, result_container = telemetry_fn(**extra_params)
    else:
        request_thread, result_container = telemetry_fn()

    assert request_thread is None
    assert result_container["response_status"] is None
    assert result_container["extra_value"] is None
    mock_urlopen.assert_not_called()


@pytest.mark.parametrize(
    "telemetry_fn,extra_params",
    [
        (track_runner_on_scarf, {"runner": "ray"}),
        (track_import_on_scarf, None),
    ],
)
@patch("daft.get_build_type")
@patch("daft.get_version")
@patch("urllib.request.urlopen")
def test_scarf_telemetry_error_handling(
    mock_urlopen: MagicMock,
    mock_version: MagicMock,
    mock_build_type: MagicMock,
    telemetry_fn,
    extra_params,
    ensure_analytics_enabled,
):
    # Test error handling in scarf_telemetry, verifies that network errors are caught, function returns error message and None for runner type

    # Clean environment
    os.environ.pop("SCARF_NO_ANALYTICS", None)
    os.environ.pop("DO_NOT_TRACK", None)

    mock_version.return_value = "0.0.0"
    mock_build_type.return_value = "release"
    mock_urlopen.side_effect = urllib.error.URLError(socket.timeout("Timeout"))

    if extra_params:
        request_thread, result_container = telemetry_fn(**extra_params)
    else:
        request_thread, result_container = telemetry_fn()

    request_thread.join()

    assert result_container["response_status"].startswith("Analytics error:")



from __future__ import annotations

# Need to import after `.table` due to circular dep issue otherwise
from .micropartition import MicroPartition
from .recordbatch import RecordBatch, read_parquet_into_pyarrow, read_parquet_into_pyarrow_bulk

__all__ = ["MicroPartition", "RecordBatch", "read_parquet_into_pyarrow", "read_parquet_into_pyarrow_bulk"]



from __future__ import annotations

import warnings


def _raw_device_count_nvml() -> int:
    """Return number of devices as reported by NVML or zero if NVML discovery/initialization failed.

    Inspired by PyTorch: https://github.com/pytorch/pytorch/blob/88e54de21976aa504e797e47f06b480b9108ef5c/torch/cuda/__init__.py#L711
    """
    from ctypes import CDLL, byref, c_int

    try:
        nvml_h = CDLL("libnvidia-ml.so.1")
    except OSError:
        return 0
    rc = nvml_h.nvmlInit()
    if rc != 0:
        warnings.warn("Can't initialize NVML, assuming no CUDA devices.")
        return 0
    dev_count = c_int(0)
    rc = nvml_h.nvmlDeviceGetCount_v2(byref(dev_count))
    if rc != 0:
        warnings.warn("Can't get nvml device count, assuming no CUDA devices.")
        return 0
    del nvml_h
    return dev_count.value


def cuda_visible_devices() -> list[str]:
    """Get the list of CUDA devices visible to the current process."""
    import os

    visible_devices_envvar = os.getenv("CUDA_VISIBLE_DEVICES")

    if visible_devices_envvar is None:
        return [str(i) for i in range(_raw_device_count_nvml())]

    return [device.strip() for device in visible_devices_envvar.split(",") if device.strip()]



from __future__ import annotations

from fastapi import FastAPI, Request, Response
from slowapi import Limiter
from slowapi.errors import RateLimitExceeded
from slowapi.util import get_remote_address

from ..utils.responses import get_response

BUCKET_NAME = "80-per-second-rate-limited-gets-bucket"
OBJECT_KEY_URL = "/{item_id}"

route = f"/{BUCKET_NAME}"
app = FastAPI()


def rate_limit_exceeded_handler(request: Request, exc: RateLimitExceeded) -> Response:
    return get_response(request.url, status_code=503, status_code_str="SlowDown")


limiter = Limiter(key_func=get_remote_address)
app.state.limiter = limiter
app.add_exception_handler(RateLimitExceeded, rate_limit_exceeded_handler)


@app.get(OBJECT_KEY_URL)
@limiter.shared_limit(limit_value="80/second", scope="my_shared_limit")
async def rate_limited_bucket_get(request: Request, item_id: str):
    """This endpoint will just echo the `item_id` and return that as the response body."""
    result = item_id.encode("utf-8")
    return Response(
        status_code=200,
        content=result,
        headers={
            "Content-Length": str(len(result)),
            "Content-Type": "binary/octet-stream",
        },
    )



from __future__ import annotations

from typing import TYPE_CHECKING

from daft.expressions.expressions import Expression, ExpressionsProjection
from daft.recordbatch.micropartition import MicroPartition

if TYPE_CHECKING:
    from daft.daft import PyExpr, PyMicroPartition

try:
    import ray
except ImportError:
    raise


@ray.remote
class UDFActor:
    def __init__(self, uninitialized_projection: ExpressionsProjection) -> None:
        self.projection = ExpressionsProjection([e._initialize_udfs() for e in uninitialized_projection])

    def get_node_id(self) -> str:
        return ray.get_runtime_context().get_node_id()

    def eval_input(self, input: PyMicroPartition) -> PyMicroPartition:
        mp = MicroPartition._from_pymicropartition(input)
        res = mp.eval_expression_list(self.projection)
        return res._micropartition


class UDFActorHandle:
    def __init__(self, node_id: str, actor_ref: ray.ObjectRef) -> None:
        self.node_id = node_id
        self.actor = actor_ref

    def eval_input(self, input: PyMicroPartition) -> PyMicroPartition:
        return ray.get(self.actor.eval_input.remote(input))

    def teardown(self) -> None:
        ray.kill(self.actor)


def start_udf_actors(
    projection: list[PyExpr],
    num_actors: int,
    num_gpus_per_actor: float,
    memory_per_actor: float,
    num_cpus_per_actor: float,
) -> list[tuple[str, list[UDFActorHandle]]]:
    expr_projection = ExpressionsProjection([Expression._from_pyexpr(expr) for expr in projection])
    handles: dict[str, list[UDFActorHandle]] = {}
    actors = [
        UDFActor.options(  # type: ignore
            scheduling_strategy="SPREAD",
            num_gpus=num_gpus_per_actor,
            num_cpus=num_cpus_per_actor,
            memory=memory_per_actor,
        ).remote(expr_projection)
        for _ in range(num_actors)
    ]
    node_ids = ray.get([actor.get_node_id.remote() for actor in actors])
    for actor, node_id in zip(actors, node_ids):
        handles.setdefault(node_id, []).append(UDFActorHandle(node_id, actor))

    res = [(node_id, handles) for node_id, handles in handles.items()]
    return res



from __future__ import annotations

import pyarrow as pa
import pytest

from daft import DataType
from daft.expressions import col, lit
from daft.recordbatch import MicroPartition


@pytest.mark.parametrize(
    "input_data,start,length,expected_result,size",
    [
        # Basic slicing
        ([b"abc", b"def", b"ghi"], [0, 1, 2], [2, 1, 1], [b"ab", b"e", b"i"], 3),
        # With nulls
        ([b"ab", None, b"cd"], [0, 1, 0], [1, 1, 2], [b"a", None, b"cd"], 2),
        # Special sequences
        ([b"\x00\x01", b"\xff\xfe", b"\xe2\x98"], [1, 0, 1], [1, 1, 1], [b"\x01", b"\xff", b"\x98"], 2),
        # Edge cases
        (
            [b"abc", b"def", b"ghi"],
            [3, 2, 1],  # Start at or beyond length
            [1, 2, 3],
            [b"", b"f", b"hi"],
            3,
        ),
        # UTF-8 sequences
        (
            [b"\xe2\x98\x83", b"\xf0\x9f\x98", b"\xf0\x9f\x8c"],  # UTF-8 characters and partials
            [0, 1, 2],
            [2, 2, 1],
            [b"\xe2\x98", b"\x9f\x98", b"\x8c"],
            3,
        ),
        # Zero bytes in different positions
        ([b"\x00ab", b"a\x00b", b"ab\x00"], [0, 1, 2], [2, 1, 1], [b"\x00a", b"\x00", b"\x00"], 3),
        # High value bytes
        ([b"\xff\xff\xff", b"\xfe\xfe\xfe", b"\xfd\xfd\xfd"], [0, 1, 2], [2, 1, 1], [b"\xff\xff", b"\xfe", b"\xfd"], 3),
        # Mixed content
        (
            [b"a\xff\x00", b"\x00\xff\x83", b"\xe2\x98\x83"],
            [1, 0, 0],
            [2, 2, 2],
            [b"\xff\x00", b"\x00\xff", b"\xe2\x98"],
            3,
        ),
    ],
)
def test_fixed_size_binary_slice(
    input_data: list[bytes | None],
    start: list[int],
    length: list[int],
    expected_result: list[bytes | None],
    size: int,
) -> None:
    table = MicroPartition.from_pydict(
        {
            "a": pa.array(input_data, type=pa.binary(size)),
            "start": start,
            "length": length,
        }
    )
    # Verify input is FixedSizeBinary before slicing
    assert table.schema()["a"].dtype == DataType.fixed_size_binary(size)
    result = table.eval_expression_list([col("a").binary.slice(col("start"), col("length"))])
    assert result.to_pydict() == {"a": expected_result}
    # Result should be regular Binary since slice might be smaller
    assert result.schema()["a"].dtype == DataType.binary()


@pytest.mark.parametrize(
    "input_data,start,expected_result,size",
    [
        # Without length parameter
        ([b"abc", b"def", b"ghi"], [1, 0, 2], [b"bc", b"def", b"i"], 3),
        # With nulls
        ([b"ab", None, b"cd"], [1, 0, 1], [b"b", None, b"d"], 2),
        # UTF-8 sequences
        ([b"\xe2\x98\x83", b"\xf0\x9f\x98", b"\xf0\x9f\x8c"], [1, 0, 2], [b"\x98\x83", b"\xf0\x9f\x98", b"\x8c"], 3),
        # Special bytes
        ([b"\x00\xff\x7f", b"\xff\x00\xff", b"\x7f\xff\x00"], [1, 2, 0], [b"\xff\x7f", b"\xff", b"\x7f\xff\x00"], 3),
    ],
)
def test_fixed_size_binary_slice_no_length(
    input_data: list[bytes | None],
    start: list[int],
    expected_result: list[bytes | None],
    size: int,
) -> None:
    table = MicroPartition.from_pydict(
        {
            "a": pa.array(input_data, type=pa.binary(size)),
            "start": start,
        }
    )
    # Verify input is FixedSizeBinary before slicing
    assert table.schema()["a"].dtype == DataType.fixed_size_binary(size)
    result = table.eval_expression_list([col("a").binary.slice(col("start"))])
    assert result.to_pydict() == {"a": expected_result}
    # Result should be regular Binary since slice might be smaller
    assert result.schema()["a"].dtype == DataType.binary()


def test_fixed_size_binary_slice_computed() -> None:
    # Test with computed start index (length - 2)
    size = 4
    table = MicroPartition.from_pydict(
        {
            "a": pa.array(
                [
                    b"abcd",  # Start at 2, take 2
                    b"\xff\xfe\xfd\xfc",  # Start at 2, take 2
                    b"\x00\x01\x02\x03",  # Start at 2, take 2
                    b"\xe2\x98\x83\x00",  # Start at 2, take 2
                ],
                type=pa.binary(size),
            ),
        }
    )
    # Verify input is FixedSizeBinary
    assert table.schema()["a"].dtype == DataType.fixed_size_binary(size)

    # Test with computed start (size - 2) and fixed length
    result = table.eval_expression_list(
        [
            col("a").binary.slice(
                (lit(size) - 2).cast(DataType.int32()),  # Start 2 chars from end
                2,  # Take 2 chars
            )
        ]
    )
    assert result.to_pydict() == {"a": [b"cd", b"\xfd\xfc", b"\x02\x03", b"\x83\x00"]}
    assert result.schema()["a"].dtype == DataType.binary()

    # Test with fixed start and computed length (size - start)
    result = table.eval_expression_list(
        [
            col("a").binary.slice(
                1,  # Start at second char
                (lit(size) - 1).cast(DataType.int32()),  # Take remaining chars
            )
        ]
    )
    assert result.to_pydict() == {"a": [b"bcd", b"\xfe\xfd\xfc", b"\x01\x02\x03", b"\x98\x83\x00"]}
    assert result.schema()["a"].dtype == DataType.binary()


def test_fixed_size_binary_slice_edge_cases() -> None:
    # Test various edge cases with different fixed sizes
    cases = [
        # Single byte values
        (
            1,  # size
            [b"\x00", b"\xff", b"a", None],  # input
            [0, 0, 0, 0],  # start
            [1, 1, 1, 1],  # length
            [b"\x00", b"\xff", b"a", None],  # expected
        ),
        # Two byte values with boundary cases
        (
            2,  # size
            [b"\x00\x01", b"\xff\xfe", b"ab", None],  # input
            [1, 2, 0, 1],  # start
            [1, 0, 2, 1],  # length
            [b"\x01", b"", b"ab", None],  # expected
        ),
        # Four byte values with various slices
        (
            4,  # size
            [b"\x00\x01\x02\x03", b"abcd", b"\xff\xfe\xfd\xfc", None],  # input
            [0, 1, 2, 0],  # start
            [2, 2, 2, 4],  # length
            [b"\x00\x01", b"bc", b"\xfd\xfc", None],  # expected
        ),
    ]

    for size, input_data, start, length, expected in cases:
        table = MicroPartition.from_pydict(
            {
                "a": pa.array(input_data, type=pa.binary(size)),
                "start": start,
                "length": length,
            }
        )
        # Verify input is FixedSizeBinary
        assert table.schema()["a"].dtype == DataType.fixed_size_binary(size)
        result = table.eval_expression_list([col("a").binary.slice(col("start"), col("length"))])
        assert result.to_pydict() == {"a": expected}
        assert result.schema()["a"].dtype == DataType.binary()


def test_fixed_size_binary_slice_with_literals() -> None:
    table = MicroPartition.from_pydict(
        {
            "a": pa.array([b"abc", b"def", None], type=pa.binary(3)),
        }
    )
    # Verify input is FixedSizeBinary before slicing
    assert table.schema()["a"].dtype == DataType.fixed_size_binary(3)

    # Test with literal start and length
    result = table.eval_expression_list([col("a").binary.slice(lit(1), lit(1))])
    assert result.to_pydict() == {"a": [b"b", b"e", None]}
    assert result.schema()["a"].dtype == DataType.binary()

    # Test with only start
    result = table.eval_expression_list([col("a").binary.slice(lit(0))])
    assert result.to_pydict() == {"a": [b"abc", b"def", None]}
    assert result.schema()["a"].dtype == DataType.binary()

    # Test with start beyond length
    result = table.eval_expression_list([col("a").binary.slice(lit(3), lit(1))])
    assert result.to_pydict() == {"a": [b"", b"", None]}
    assert result.schema()["a"].dtype == DataType.binary()

    # Test with zero length
    result = table.eval_expression_list([col("a").binary.slice(lit(0), lit(0))])
    assert result.to_pydict() == {"a": [b"", b"", None]}
    assert result.schema()["a"].dtype == DataType.binary()


def test_fixed_size_binary_slice_errors() -> None:
    # Test error cases
    table = MicroPartition.from_pydict(
        {
            "a": pa.array([b"abc", b"def"], type=pa.binary(3)),
            "b": [1, 2],  # Wrong type
            "start": [-1, 0],  # Negative start
            "length": [0, -1],  # Negative length
        }
    )

    # Test slice on wrong type
    with pytest.raises(
        Exception, match="Expects inputs to binary_slice to be binary, integer and integer or null but received Int64"
    ):
        table.eval_expression_list([col("b").binary.slice(lit(0))])

    # Test negative start
    with pytest.raises(Exception, match="DaftError::ComputeError Failed to cast numeric value to target type"):
        table.eval_expression_list([col("a").binary.slice(col("start"))])

    # Test negative length
    with pytest.raises(Exception, match="DaftError::ComputeError Failed to cast numeric value to target type"):
        table.eval_expression_list([col("a").binary.slice(lit(0), col("length"))])

    # Test with wrong number of arguments (too many)
    with pytest.raises(
        Exception,
        match="(?:ExpressionBinaryNamespace.)?slice\\(\\) takes from 2 to 3 positional arguments but 4 were given",
    ):
        table.eval_expression_list([col("a").binary.slice(lit(0), lit(1), lit(2))])

    # Test with wrong number of arguments (too few)
    with pytest.raises(
        Exception, match="(?:ExpressionBinaryNamespace.)?slice\\(\\) missing 1 required positional argument: 'start'"
    ):
        table.eval_expression_list([col("a").binary.slice()])



from __future__ import annotations

from daft.pickle.cloudpickle_fast import dumps as cloudpickle_dumps
from daft.pickle.cloudpickle_fast import loads as cloudpickle_loads


def dumps(obj: object) -> bytes:
    return cloudpickle_dumps(obj)


def loads(data: bytes) -> object:
    return cloudpickle_loads(data)



from __future__ import annotations

from flytekit import current_context, task, workflow
from flytekitplugins.ray import (  # noqa # pylint: disable=unused-import
    RayJobConfig,
    WorkerNodeConfig,
)

import daft


###
# Run Daft on a Ray cluster
###
# @task(
#     task_config=RayJobConfig(
#         worker_node_config=[
#             WorkerNodeConfig(
#                 group_name="ray-group", replicas=1,
#             )
#         ],
#     )
# )
#
###
# Run Daft on a (big!) Flyte task
###
@task()
def produce_resized_image_dataset(limit: int) -> list[str]:
    # NOTE: Use Ray Runner:
    #
    #     1. If Ray connection has been set up already by Flyte, it will use the initialized Ray cluster connection
    #     2. Otherwise, it will create a local Ray cluster on the Flyte task
    #
    daft.context.set_runner_ray()

    written_df = daft_notebook_code(limit)

    # Return the number of rows written
    return written_df.to_pydict()["file_path"]


@workflow()
def wf(limit: int):
    produce_resized_image_dataset(limit=limit)


def daft_notebook_code(limit: int):
    # Download data from Parquet file
    IO_CONFIG = IO_CONFIG = daft.io.IOConfig(
        s3=daft.io.S3Config(anonymous=True, region_name="us-west-2")
    )  # Use anonymous-mode for accessing AWS S3
    PARQUET_PATH = "s3://daft-public-data/tutorials/laion-parquet/train-00000-of-00001-6f24a7497df494ae.parquet"
    parquet_df = daft.read_parquet(PARQUET_PATH, io_config=IO_CONFIG)
    parquet_df = parquet_df.select(parquet_df["URL"], parquet_df["TEXT"], parquet_df["AESTHETIC_SCORE"])

    # Filter for only "pretty" images
    filtered_df = parquet_df.where(parquet_df["TEXT"].str.contains("darkness"))

    # Download and resize images
    filtered_df = filtered_df.with_column(
        "image",
        filtered_df["URL"].url.download(on_error="null").image.decode(),
    )
    filtered_df = filtered_df.with_column(
        "resized_image",
        filtered_df["image"].image.resize(32, 32),
    )

    # Write processed data out to Parquet file
    written_df = (
        filtered_df.select("URL", "TEXT", "resized_image")
        .limit(limit)
        .write_parquet(
            # NOTE: In practice, you will be writing to cloud storage such as AWS S3.
            # This currently writes to a local (ephemeral) file.
            f"my-s3-bucket/{current_context().execution_id.name}/resized_images.parquet"
        )
    )
    return written_df



from __future__ import annotations

import tempfile

import boto3
import pytest
from botocore import UNSIGNED
from botocore.client import Config

import daft
from daft.io import IOConfig, S3Config


@pytest.fixture
def expected_count(request):
    return request.param


@pytest.fixture
def io_config():
    return IOConfig(s3=S3Config(anonymous=True))


@pytest.fixture(scope="session")
def files(request):
    num_files = request.param
    s3 = boto3.client("s3", config=Config(signature_version=UNSIGNED))
    with tempfile.TemporaryDirectory() as tmpdir:
        local_file = f"{tmpdir}/small-fake-data.parquet.parquet"

        s3.download_file("daft-public-data", "test_fixtures/parquet/small-fake-data.parquet", local_file)
        if request.param == 1:
            yield local_file
        else:
            yield [local_file] * num_files


@pytest.mark.benchmark(group="show")
@pytest.mark.parametrize(
    "files",
    [
        pytest.param(1, id="1 Small File"),
        pytest.param(100, id="100 Small Files"),
    ],
    indirect=True,  # This tells pytest to pass the params to the fixture
)
def test_show(files, io_config, benchmark):
    def f():
        df = daft.read_parquet(files, io_config=io_config)
        df.show()

    benchmark(f)


@pytest.mark.benchmark(group="explain")
@pytest.mark.parametrize(
    "files",
    [
        pytest.param(1, id="1 Small File"),
        pytest.param(100, id="100 Small Files"),
    ],
    indirect=True,  # This tells pytest to pass the params to the fixture
)
def test_explain(files, io_config, benchmark):
    def f():
        df = daft.read_parquet(files, io_config=io_config)
        df.explain(True)

    benchmark(f)


@pytest.mark.benchmark(group="count")
@pytest.mark.parametrize(
    "files, expected_count",
    [
        pytest.param(1, 1024, id="1 Small File"),
        pytest.param(100, 102400, id="100 Small Files"),
    ],
    indirect=True,  # This tells pytest to pass the params to the fixture
)
def test_count(files, expected_count, io_config, benchmark):
    def f():
        df = daft.read_parquet(files, io_config=io_config)
        return df.count_rows()

    count = benchmark(f)
    assert count == expected_count


@pytest.mark.benchmark(group="iter_rows")
@pytest.mark.parametrize(
    "files",
    [
        pytest.param(1, id="1 Small File"),
        pytest.param(100, id="100 Small Files"),
    ],
    indirect=True,  # This tells pytest to pass the params to the fixture
)
def test_iter_rows_first_row(files, io_config, benchmark):
    def f():
        df = daft.read_parquet(files, io_config=io_config)
        iter = df.iter_rows()
        return next(iter)

    first_row = benchmark(f)
    assert len(first_row.keys()) == 16



from __future__ import annotations

import pyarrow as pa
import pyarrow.parquet as papq
import pytest

from daft.daft import testing as native_testing_utils
from daft.recordbatch.micropartition import MicroPartition

pytest.skip(allow_module_level=True, reason="Skipping because these tests don't currently pass")


def get_scantask_estimated_size(pq_path: str, size_on_disk: int, columns: list[str] | None = None) -> int:
    """Retrieve the estimated size for reading a given Parquet file."""
    return native_testing_utils.estimate_in_memory_size_bytes(str(pq_path), size_on_disk, columns=columns)


def get_actual_size(pq_path: str, columns: list[str] | None = None) -> int:
    # Force materializationm of the MicroPartition using `.slice`
    mp = MicroPartition.read_parquet(str(pq_path), columns=columns)
    return mp.slice(0, len(mp)).size_bytes()


def assert_close(size_on_disk, estimated: int, actual: int, pct: float = 0.3):
    assert (
        abs(actual - estimated) / estimated < pct
    ), f"Expected {size_on_disk / 1_000_000:.2f}MB file estimated vs actual to be within {pct * 100}%: {estimated / 1_000_000:.2f}MB vs {actual / 1000_000:.2f}MB ({((estimated - actual) / estimated) * 100:.2f}%)"


@pytest.mark.parametrize("compression", ["snappy", None], ids=["snappy", "no_compression"])
@pytest.mark.parametrize("use_dictionary", [True, False], ids=["dict", "no_dict"])
def test_estimations_unique_strings(tmpdir, use_dictionary, compression):
    pq_path = tmpdir / f"unique_strings.use_dictionary={use_dictionary}.compression={compression}.pq"
    data = [f"{'a' * 100}{i}" for i in range(1000_000)]
    tbl = pa.table({"foo": data})
    papq.write_table(tbl, pq_path, use_dictionary=use_dictionary, compression=compression)

    size_on_disk = pq_path.stat().size
    assert_close(size_on_disk, get_scantask_estimated_size(pq_path, size_on_disk), get_actual_size(pq_path))


@pytest.mark.parametrize("compression", ["snappy", None], ids=["snappy", "no_compression"])
@pytest.mark.parametrize("use_dictionary", [True, False], ids=["dict", "no_dict"])
def test_estimations_dup_strings(tmpdir, use_dictionary, compression):
    pq_path = tmpdir / f"dup_strings.use_dictionary={use_dictionary}.compression={compression}.pq"
    data = ["a" * 100 for _ in range(1000_000)]
    tbl = pa.table({"foo": data})
    papq.write_table(tbl, pq_path, use_dictionary=use_dictionary, compression=compression)

    size_on_disk = pq_path.stat().size
    assert_close(size_on_disk, get_scantask_estimated_size(pq_path, size_on_disk), get_actual_size(pq_path))


@pytest.mark.parametrize("compression", ["snappy", None], ids=["snappy", "no_compression"])
@pytest.mark.parametrize("use_dictionary", [True, False], ids=["dict", "no_dict"])
def test_estimations_unique_ints(tmpdir, use_dictionary, compression):
    pq_path = tmpdir / f"unique_ints.use_dictionary={use_dictionary}.compression={compression}.pq"

    data = [i for i in range(1000_000)]
    tbl = pa.table({"foo": data})
    papq.write_table(tbl, pq_path, use_dictionary=use_dictionary, compression=compression)

    size_on_disk = pq_path.stat().size
    assert_close(size_on_disk, get_scantask_estimated_size(pq_path, size_on_disk), get_actual_size(pq_path))


@pytest.mark.parametrize("compression", ["snappy", None], ids=["snappy", "no_compression"])
@pytest.mark.parametrize("use_dictionary", [True, False], ids=["dict", "no_dict"])
def test_estimations_dup_ints(tmpdir, use_dictionary, compression):
    pq_path = tmpdir / f"dup_ints.use_dictionary={use_dictionary}.compression={compression}.pq"

    data = [1 for _ in range(1000_000)]
    tbl = pa.table({"foo": data})
    papq.write_table(tbl, pq_path, use_dictionary=use_dictionary, compression=compression)

    size_on_disk = pq_path.stat().size
    assert_close(size_on_disk, get_scantask_estimated_size(pq_path, size_on_disk), get_actual_size(pq_path))


@pytest.mark.parametrize(
    "path",
    [
        "https://huggingface.co/datasets/HuggingFaceTB/smoltalk/resolve/main/data/all/test-00000-of-00001.parquet",
        "https://huggingface.co/datasets/uoft-cs/cifar10/resolve/main/plain_text/test-00000-of-00001.parquet",
        "https://huggingface.co/datasets/stanfordnlp/imdb/resolve/main/plain_text/unsupervised-00000-of-00001.parquet",
        "https://huggingface.co/datasets/jat-project/jat-dataset-tokenized/resolve/main/atari-alien/test-00000-of-00011.parquet",
        "https://huggingface.co/datasets/princeton-nlp/SWE-bench_Verified/resolve/main/data/test-00000-of-00001.parquet",
        "https://huggingface.co/datasets/lmms-lab/LLaVA-OneVision-Data/resolve/main/CLEVR-Math(MathV360K)/train-00000-of-00002.parquet",
    ],
    ids=[
        "smoktalk",
        "cifar10",
        "standfordnlp-imdb",
        "jat-dataset-tokenized",
        "swe-bench-verified",
        "llava-onevision-data",
    ],
)
def test_canonical_files_in_hf(path):
    import requests

    response = requests.head(path, allow_redirects=True)
    size_on_disk = int(response.headers["Content-Length"])

    assert_close(size_on_disk, get_scantask_estimated_size(path, size_on_disk), get_actual_size(path))


@pytest.mark.parametrize(
    "path",
    [
        "s3://daft-public-datasets/tpch_iceberg_sf1000.db/lineitem/data/L_SHIPDATE_month=1992-01/00000-6694-fa4594d5-f624-407c-8640-5b6db8150470-00001.parquet",
    ],
    ids=[
        "lineitem",
    ],
)
def test_canonical_files_in_s3(path):
    import boto3

    s3 = boto3.client("s3")
    bucket, key = path.replace("s3://", "").split("/", 1)
    response = s3.head_object(Bucket=bucket, Key=key)
    size_on_disk = response["ContentLength"]

    assert_close(size_on_disk, get_scantask_estimated_size(path, size_on_disk), get_actual_size(path))



from __future__ import annotations

import adlfs
import pytest

from daft.daft import AzureConfig, IOConfig, io_glob

STORAGE_ACCOUNT = "dafttestdata"
CONTAINER = "public-anonymous"
DEFAULT_AZURE_CONFIG = AzureConfig(storage_account=STORAGE_ACCOUNT, anonymous=True)


def adlfs_recursive_list(fs, path) -> list:
    all_results = []
    curr_level_result = fs.ls(path.replace("az://", ""), detail=True)
    for item in curr_level_result:
        if item["type"] == "directory":
            new_path = f'az://{item["name"]}'
            all_results.extend(adlfs_recursive_list(fs, new_path))
            item["name"] += "/"
            all_results.append(item)
        else:
            all_results.append(item)
    return all_results


def compare_az_result(daft_ls_result: list, fsspec_result: list):
    daft_files = [(f["path"], f["type"].lower()) for f in daft_ls_result]
    azfs_files = [(f"az://{f['name']}", f["type"]) for f in fsspec_result]

    # Remove all directories: our glob utilities don't return dirs
    azfs_files = [(path, type_) for path, type_ in azfs_files if type_ == "file"]

    assert len(daft_files) == len(azfs_files)
    assert sorted(daft_files) == sorted(azfs_files)


@pytest.mark.integration()
@pytest.mark.parametrize(
    "path",
    [
        f"az://{CONTAINER}",
        f"az://{CONTAINER}/",
        f"az://{CONTAINER}/test_ls/",
        f"az://{CONTAINER}/test_ls//",
    ],
)
@pytest.mark.parametrize("recursive", [False, True])
@pytest.mark.parametrize("fanout_limit", [None, 1])
def test_az_flat_directory_listing(path, recursive, fanout_limit):
    fs = adlfs.AzureBlobFileSystem(account_name=STORAGE_ACCOUNT)
    glob_path = path.rstrip("/") + "/**/*.*" if recursive else path
    daft_ls_result = io_glob(glob_path, io_config=IOConfig(azure=DEFAULT_AZURE_CONFIG), fanout_limit=fanout_limit)
    fsspec_result = adlfs_recursive_list(fs, path) if recursive else fs.ls(path.replace("az://", ""), detail=True)
    compare_az_result(daft_ls_result, fsspec_result)


@pytest.mark.integration()
def test_az_single_file_listing():
    path = f"az://{CONTAINER}/mvp.parquet"
    fs = adlfs.AzureBlobFileSystem(account_name=STORAGE_ACCOUNT)
    daft_ls_result = io_glob(path, io_config=IOConfig(azure=DEFAULT_AZURE_CONFIG))
    fsspec_result = fs.ls(path.replace("az://", ""), detail=True)
    compare_az_result(daft_ls_result, fsspec_result)


@pytest.mark.integration()
def test_az_notfound():
    path = f"az://{CONTAINER}/test_"
    with pytest.raises(FileNotFoundError, match=path):
        io_glob(path, io_config=IOConfig(azure=DEFAULT_AZURE_CONFIG))



from __future__ import annotations

import functools
from typing import TYPE_CHECKING, Any, Callable

from daft.context import get_context
from daft.daft import (
    CountMode,
    FileFormat,
    IOConfig,
    JoinStrategy,
    JoinType,
    PyDaftExecutionConfig,
    ScanOperatorHandle,
    WriteMode,
    logical_plan_table_scan,
)
from daft.daft import LogicalPlanBuilder as _LogicalPlanBuilder
from daft.expressions import Expression, col
from daft.logical.schema import Schema

if TYPE_CHECKING:
    import pathlib

    from pyiceberg.table import Table as IcebergTable

    from daft.io import DataSink
    from daft.plan_scheduler.physical_plan_scheduler import (
        AdaptivePhysicalPlanScheduler,
        PhysicalPlanScheduler,
    )
    from daft.runners.partitioning import PartitionCacheEntry


def _apply_daft_planning_config_to_initializer(
    classmethod_func: Callable[..., LogicalPlanBuilder],
) -> Callable[..., LogicalPlanBuilder]:
    """Decorator to be applied to any @classmethod instantiation method on LogicalPlanBuilder.

    This decorator ensures that the current DaftPlanningConfig is applied to the instantiated LogicalPlanBuilder
    """

    @functools.wraps(classmethod_func)
    def wrapper(cls: type[LogicalPlanBuilder], *args: Any, **kwargs: Any) -> LogicalPlanBuilder:
        instantiated_logical_plan_builder = classmethod_func(cls, *args, **kwargs)

        # Parametrize the builder with the current DaftPlanningConfig
        inner = instantiated_logical_plan_builder._builder
        inner = inner.with_planning_config(get_context().daft_planning_config)

        return cls(inner)

    return wrapper


class LogicalPlanBuilder:
    """A logical plan builder for the Daft DataFrame."""

    def __init__(self, builder: _LogicalPlanBuilder) -> None:
        self._builder = builder

    def to_physical_plan_scheduler(self, daft_execution_config: PyDaftExecutionConfig) -> PhysicalPlanScheduler:
        """Convert the underlying logical plan to a physical plan scheduler.

        physical plan scheduler is used to generate executable tasks for the physical plan.

        This should be called after triggering optimization with self.optimize().

        **Warning**: This function is not part of the stable API and may change
        without notice. It is intended for internal or experimental use only.
        """
        from daft.plan_scheduler.physical_plan_scheduler import PhysicalPlanScheduler

        return PhysicalPlanScheduler.from_logical_plan_builder(
            self,
            daft_execution_config,
        )

    def to_adaptive_physical_plan_scheduler(
        self, daft_execution_config: PyDaftExecutionConfig
    ) -> AdaptivePhysicalPlanScheduler:
        from daft.plan_scheduler.physical_plan_scheduler import (
            AdaptivePhysicalPlanScheduler,
        )

        return AdaptivePhysicalPlanScheduler.from_logical_plan_builder(
            self,
            daft_execution_config,
        )

    def schema(self) -> Schema:
        """The schema of the current logical plan."""
        pyschema = self._builder.schema()
        return Schema._from_pyschema(pyschema)

    def describe(self) -> LogicalPlanBuilder:
        builder = self._builder.describe()
        return LogicalPlanBuilder(builder)

    def summarize(self) -> LogicalPlanBuilder:
        builder = self._builder.summarize()
        return LogicalPlanBuilder(builder)

    def pretty_print(self, simple: bool = False, format: str = "ascii") -> str:
        """Pretty prints the current underlying logical plan."""
        from daft.dataframe.display import MermaidOptions

        if format == "ascii":
            return self._builder.repr_ascii(simple)
        elif format == "mermaid":
            return self._builder.repr_mermaid(MermaidOptions(simple))
        else:
            raise ValueError(f"Unknown format: {format}")

    def __repr__(self) -> str:
        return self._builder.repr_ascii(simple=False)

    def optimize(self) -> LogicalPlanBuilder:
        """Optimize the underlying logical plan."""
        builder = self._builder.optimize()
        return LogicalPlanBuilder(builder)

    @classmethod
    @_apply_daft_planning_config_to_initializer
    def from_in_memory_scan(
        cls,
        cache_entry: PartitionCacheEntry,
        schema: Schema,
        num_partitions: int,
        size_bytes: int,
        num_rows: int,
    ) -> LogicalPlanBuilder:
        builder = _LogicalPlanBuilder.in_memory_scan(
            cache_entry.key,
            cache_entry,
            schema._schema,
            num_partitions,
            size_bytes,
            num_rows,
        )
        return cls(builder)

    @classmethod
    @_apply_daft_planning_config_to_initializer
    def from_tabular_scan(
        cls,
        *,
        scan_operator: ScanOperatorHandle,
    ) -> LogicalPlanBuilder:
        builder = logical_plan_table_scan(scan_operator)
        return cls(builder)

    def select(
        self,
        to_select: list[Expression],
    ) -> LogicalPlanBuilder:
        to_select_pyexprs = [expr._expr for expr in to_select]
        builder = self._builder.select(to_select_pyexprs)
        return LogicalPlanBuilder(builder)

    def with_columns(self, columns: list[Expression]) -> LogicalPlanBuilder:
        column_pyexprs = [expr._expr for expr in columns]
        builder = self._builder.with_columns(column_pyexprs)
        return LogicalPlanBuilder(builder)

    def with_column_renamed(self, existing: str, new: str) -> LogicalPlanBuilder:
        cols_map = {existing: new}
        builder = self._builder.with_columns_renamed(cols_map)
        return LogicalPlanBuilder(builder)

    def with_columns_renamed(self, cols_map: dict[str, str]) -> LogicalPlanBuilder:
        builder = self._builder.with_columns_renamed(cols_map)
        return LogicalPlanBuilder(builder)

    def exclude(self, to_exclude: list[str]) -> LogicalPlanBuilder:
        builder = self._builder.exclude(to_exclude)
        return LogicalPlanBuilder(builder)

    def filter(self, predicate: Expression) -> LogicalPlanBuilder:
        builder = self._builder.filter(predicate._expr)
        return LogicalPlanBuilder(builder)

    def limit(self, num_rows: int, eager: bool) -> LogicalPlanBuilder:
        builder = self._builder.limit(num_rows, eager)
        return LogicalPlanBuilder(builder)

    def explode(self, explode_expressions: list[Expression]) -> LogicalPlanBuilder:
        explode_pyexprs = [expr._expr for expr in explode_expressions]
        builder = self._builder.explode(explode_pyexprs)
        return LogicalPlanBuilder(builder)

    def unpivot(
        self,
        ids: list[Expression],
        values: list[Expression],
        variable_name: str,
        value_name: str,
    ) -> LogicalPlanBuilder:
        ids_pyexprs = [expr._expr for expr in ids]
        values_pyexprs = [expr._expr for expr in values]
        builder = self._builder.unpivot(ids_pyexprs, values_pyexprs, variable_name, value_name)
        return LogicalPlanBuilder(builder)

    def count(self) -> LogicalPlanBuilder:
        # TODO(Clark): Add dedicated logical/physical ops when introducing metadata-based count optimizations.
        first_col = col(self.schema().column_names()[0])
        builder = self._builder.aggregate([first_col.count(CountMode.All)._expr], [])
        builder = builder.select([first_col.alias("count")._expr])
        return LogicalPlanBuilder(builder)

    def distinct(self) -> LogicalPlanBuilder:
        builder = self._builder.distinct()
        return LogicalPlanBuilder(builder)

    def sample(self, fraction: float, with_replacement: bool, seed: int | None) -> LogicalPlanBuilder:
        builder = self._builder.sample(fraction, with_replacement, seed)
        return LogicalPlanBuilder(builder)

    def sort(
        self,
        sort_by: list[Expression],
        descending: list[bool] | bool = False,
        nulls_first: list[bool] | bool | None = None,
    ) -> LogicalPlanBuilder:
        sort_by_pyexprs = [expr._expr for expr in sort_by]
        if not isinstance(descending, list):
            descending = [descending] * len(sort_by_pyexprs)
        if nulls_first is None:
            nulls_first = descending
        elif isinstance(nulls_first, bool):
            nulls_first = [nulls_first] * len(sort_by_pyexprs)
        builder = self._builder.sort(sort_by_pyexprs, descending, nulls_first)
        return LogicalPlanBuilder(builder)

    def hash_repartition(self, num_partitions: int | None, partition_by: list[Expression]) -> LogicalPlanBuilder:
        partition_by_pyexprs = [expr._expr for expr in partition_by]
        builder = self._builder.hash_repartition(partition_by_pyexprs, num_partitions=num_partitions)
        return LogicalPlanBuilder(builder)

    def random_shuffle(self, num_partitions: int | None) -> LogicalPlanBuilder:
        builder = self._builder.random_shuffle(num_partitions)
        return LogicalPlanBuilder(builder)

    def into_partitions(self, num_partitions: int) -> LogicalPlanBuilder:
        builder = self._builder.into_partitions(num_partitions)
        return LogicalPlanBuilder(builder)

    def agg(
        self,
        to_agg: list[Expression],
        group_by: list[Expression] | None,
    ) -> LogicalPlanBuilder:
        group_by_pyexprs = [expr._expr for expr in group_by] if group_by is not None else []
        builder = self._builder.aggregate([expr._expr for expr in to_agg], group_by_pyexprs)
        return LogicalPlanBuilder(builder)

    def map_groups(self, udf: Expression, group_by: list[Expression] | None) -> LogicalPlanBuilder:
        group_by_pyexprs = [expr._expr for expr in group_by] if group_by is not None else []
        builder = self._builder.aggregate([udf._expr], group_by_pyexprs)
        return LogicalPlanBuilder(builder)

    def pivot(
        self,
        group_by: list[Expression],
        pivot_col: Expression,
        value_col: Expression,
        agg_fn: Expression,
        names: list[str],
    ) -> LogicalPlanBuilder:
        group_by_pyexprs = [expr._expr for expr in group_by]
        builder = self._builder.pivot(group_by_pyexprs, pivot_col._expr, value_col._expr, agg_fn._expr, names)
        return LogicalPlanBuilder(builder)

    def join(
        self,
        right: LogicalPlanBuilder,
        left_on: list[Expression],
        right_on: list[Expression],
        how: JoinType = JoinType.Inner,
        strategy: JoinStrategy | None = None,
        prefix: str | None = None,
        suffix: str | None = None,
    ) -> LogicalPlanBuilder:
        builder = self._builder.join(
            right._builder,
            [expr._expr for expr in left_on],
            [expr._expr for expr in right_on],
            how,
            strategy,
            prefix,
            suffix,
        )
        return LogicalPlanBuilder(builder)

    def concat(self, other: LogicalPlanBuilder) -> LogicalPlanBuilder:
        builder = self._builder.concat(other._builder)
        return LogicalPlanBuilder(builder)

    def union(self, other: LogicalPlanBuilder, is_all: bool = False, is_by_name: bool = False) -> LogicalPlanBuilder:
        builder = self._builder.union(other._builder, is_all, is_by_name)
        return LogicalPlanBuilder(builder)

    def intersect(self, other: LogicalPlanBuilder) -> LogicalPlanBuilder:
        builder = self._builder.intersect(other._builder, False)
        return LogicalPlanBuilder(builder)

    def intersect_all(self, other: LogicalPlanBuilder) -> LogicalPlanBuilder:
        builder = self._builder.intersect(other._builder, True)
        return LogicalPlanBuilder(builder)

    def except_distinct(self, other: LogicalPlanBuilder) -> LogicalPlanBuilder:
        builder = self._builder.except_(other._builder, False)
        return LogicalPlanBuilder(builder)

    def except_all(self, other: LogicalPlanBuilder) -> LogicalPlanBuilder:
        builder = self._builder.except_(other._builder, True)
        return LogicalPlanBuilder(builder)

    def add_monotonically_increasing_id(self, column_name: str | None) -> LogicalPlanBuilder:
        builder = self._builder.add_monotonically_increasing_id(column_name)
        return LogicalPlanBuilder(builder)

    def write_tabular(
        self,
        root_dir: str | pathlib.Path,
        write_mode: WriteMode,
        file_format: FileFormat,
        io_config: IOConfig,
        partition_cols: list[Expression] | None = None,
        compression: str | None = None,
    ) -> LogicalPlanBuilder:
        if file_format != FileFormat.Csv and file_format != FileFormat.Parquet:
            raise ValueError(f"Writing is only supported for Parquet and CSV file formats, but got: {file_format}")
        part_cols_pyexprs = [expr._expr for expr in partition_cols] if partition_cols is not None else None
        builder = self._builder.table_write(
            str(root_dir), write_mode, file_format, part_cols_pyexprs, compression, io_config
        )
        return LogicalPlanBuilder(builder)

    def write_iceberg(self, table: IcebergTable, io_config: IOConfig) -> LogicalPlanBuilder:
        from daft.iceberg.iceberg_write import get_missing_columns, partition_field_to_expr

        name = ".".join(table.name())
        location = f"{table.location()}/data"
        partition_spec = table.spec()
        schema = table.schema()
        missing_columns = get_missing_columns(self.schema().to_pyarrow_schema(), schema)
        builder = (
            self._builder
            if len(missing_columns) == 0
            else self._builder.with_columns([c._expr for c in missing_columns])
        )
        partition_cols = [partition_field_to_expr(field, schema)._expr for field in partition_spec.fields]
        props = table.properties
        columns = [col.name for col in schema.columns]
        builder = builder.iceberg_write(
            name, location, partition_spec.spec_id, partition_cols, schema, props, columns, io_config
        )
        return LogicalPlanBuilder(builder)

    def write_deltalake(
        self,
        path: str | pathlib.Path,
        mode: str,
        version: int,
        large_dtypes: bool,
        io_config: IOConfig,
        partition_cols: list[str] | None = None,
    ) -> LogicalPlanBuilder:
        columns_name = self.schema().column_names()
        builder = self._builder.delta_write(
            str(path),
            columns_name,
            mode,
            version,
            large_dtypes,
            partition_cols,
            io_config,
        )
        return LogicalPlanBuilder(builder)

    def write_lance(
        self,
        path: str | pathlib.Path,
        mode: str,
        io_config: IOConfig,
        kwargs: dict[str, Any] | None,
    ) -> LogicalPlanBuilder:
        columns_name = self.schema().column_names()
        builder = self._builder.lance_write(
            str(path),
            columns_name,
            mode,
            io_config,
            kwargs,
        )
        return LogicalPlanBuilder(builder)

    def write_datasink(self, name: str, sink: DataSink[Any]) -> LogicalPlanBuilder:
        builder = self._builder.datasink_write(name, sink)
        return LogicalPlanBuilder(builder)



from __future__ import annotations

import abc
from typing import TYPE_CHECKING

from daft.daft import (
    PyPartitionField,
    PyPartitionTransform,
    PyPushdowns,
    ScanTask,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from daft.logical.schema import Field, Schema


def make_partition_field(
    field: Field, source_field: Field | None = None, transform: PyPartitionTransform | None = None
) -> PyPartitionField:
    return PyPartitionField(
        field._field,
        source_field._field if source_field is not None else None,
        transform,
    )


class ScanOperator(abc.ABC):
    """ScanOperator is the legacy python DataSource ABC and is being migrated to daft.io.source.DataSource.

    In Daft 0.5.0 we will change the pushdown parameter from daft.daft.Pushdowns to
    daft.io.Pushdowns. For now, please use `Pushdowns._from_pypushdowns(py_pushdowns)`
    to convert the rust expressions to this python pushdowns class.
    """

    @abc.abstractmethod
    def schema(self) -> Schema:
        """Returns the schema of the data source."""
        raise NotImplementedError()

    @abc.abstractmethod
    def display_name(self) -> str:
        """Returns a human-readable name for this scan operator."""
        return self.__class__.__name__

    @abc.abstractmethod
    def partitioning_keys(self) -> list[PyPartitionField]:
        """Returns the partitioning keys for this data source."""
        raise NotImplementedError()

    @abc.abstractmethod
    def can_absorb_filter(self) -> bool:
        """Returns true if this scan can accept predicate pushdowns."""
        raise NotImplementedError()

    @abc.abstractmethod
    def can_absorb_limit(self) -> bool:
        """Returns true if this scan can accept limit pushdowns."""
        raise NotImplementedError()

    @abc.abstractmethod
    def can_absorb_select(self) -> bool:
        """Returns true if this scan can accept projection pushdowns."""
        raise NotImplementedError()

    @abc.abstractmethod
    def multiline_display(self) -> list[str]:
        """Returns a multi-line string representation of this scan operator."""
        raise NotImplementedError()

    @abc.abstractmethod
    def to_scan_tasks(self, pushdowns: PyPushdowns) -> Iterator[ScanTask]:
        """Converts this scan operator into scan tasks with the given pushdowns."""
        raise NotImplementedError()



from __future__ import annotations

import pyarrow as pa
import pytest

from daft.expressions import col
from daft.recordbatch import MicroPartition
from daft.series import Series

TEST_DATA = [
    Series.from_arrow(pa.array([[1, 2], [3, 4], None, []], type=pa.list_(pa.int64()))),
    Series.from_arrow(pa.array([[1, 2], [3, 4], None, []], type=pa.large_list(pa.int64()))),
    Series.from_arrow(pa.array([[1, 2], [3, 4], None, None], type=pa.list_(pa.int64(), list_size=2))),
]


@pytest.mark.parametrize(
    "data",
    TEST_DATA,
)
def test_explode(data):
    table = MicroPartition.from_pydict({"nested": data, "sidecar": ["a", "b", "c", "d"]})
    table = table.explode([col("nested")._explode()])
    assert table.column_names() == ["nested", "sidecar"]
    assert table.to_pydict() == {"nested": [1, 2, 3, 4, None, None], "sidecar": ["a", "a", "b", "b", "c", "d"]}


@pytest.mark.parametrize(
    "data",
    TEST_DATA,
)
def test_explode_flipped(data):
    table = MicroPartition.from_pydict({"sidecar": ["a", "b", "c", "d"], "nested": data})
    table = table.explode([col("nested")._explode()])
    assert table.column_names() == ["sidecar", "nested"]
    assert table.to_pydict() == {"nested": [1, 2, 3, 4, None, None], "sidecar": ["a", "a", "b", "b", "c", "d"]}


@pytest.mark.parametrize(
    "data",
    TEST_DATA,
)
def test_explode_multiple_cols(data):
    table = MicroPartition.from_pydict({"nested": data, "nested2": data, "sidecar": ["a", "b", "c", "d"]})
    table = table.explode([col("nested")._explode(), col("nested2")._explode()])
    assert table.column_names() == ["nested", "nested2", "sidecar"]
    assert table.to_pydict() == {
        "nested": [1, 2, 3, 4, None, None],
        "nested2": [1, 2, 3, 4, None, None],
        "sidecar": ["a", "a", "b", "b", "c", "d"],
    }


def test_explode_multiple_cols_mixed_types():
    data1 = pa.array([[1, 2], [3, 4], None, None], type=pa.list_(pa.int64()))
    data2 = pa.array([[1, 2], [3, 4], None, None], type=pa.list_(pa.int64(), list_size=2))
    table = MicroPartition.from_pydict({"nested": data1, "nested2": data2, "sidecar": ["a", "b", "c", "d"]})
    table = table.explode([col("nested")._explode(), col("nested2")._explode()])
    assert table.to_pydict() == {
        "nested": [1, 2, 3, 4, None, None],
        "nested2": [1, 2, 3, 4, None, None],
        "sidecar": ["a", "a", "b", "b", "c", "d"],
    }


def test_explode_bad_multiple_cols():
    table = MicroPartition.from_pydict(
        {
            "nested": [[1, 2, 3], [4], None, None],
            "nested2": [[1, 2], [3, 4], None, None],
            "sidecar": ["a", "b", "c", "d"],
        }
    )
    with pytest.raises(ValueError, match="In multicolumn explode, list length did not match"):
        table.explode([col("nested")._explode(), col("nested2")._explode()])


@pytest.mark.parametrize(
    "data",
    TEST_DATA,
)
def test_explode_multiple_cols_with_alias(data):
    table = MicroPartition.from_pydict({"nested": data, "nested2": data, "sidecar": ["a", "b", "c", "d"]})
    table = table.explode([col("nested").alias("nested3")._explode(), col("nested2")._explode()])
    assert table.column_names() == ["nested", "nested2", "sidecar", "nested3"]
    data_py = data.to_pylist()
    assert table.to_pydict() == {
        "nested": [data_py[0], data_py[0], data_py[1], data_py[1], data_py[2], data_py[3]],
        "nested2": [1, 2, 3, 4, None, None],
        "nested3": [1, 2, 3, 4, None, None],
        "sidecar": ["a", "a", "b", "b", "c", "d"],
    }


@pytest.mark.parametrize(
    "data",
    TEST_DATA,
)
def test_explode_eval_expr(data):
    table = MicroPartition.from_pydict({"nested": data})
    table = table.eval_expression_list([col("nested")._explode()])
    assert table.to_pydict() == {"nested": [1, 2, 3, 4, None, None]}


def test_explode_bad_col_type():
    table = MicroPartition.from_pydict({"a": [1, 2, 3]})
    with pytest.raises(ValueError, match="to be a List Type, but is"):
        table = table.explode([col("a")._explode()])



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from daft.expressions import Expression

if TYPE_CHECKING:
    from daft.daft import PyPushdowns


@dataclass(frozen=True)
class Pushdowns:
    """Pushdowns are sent to scan sources during query planning.

    Attributes:
        filters (Expression | None): Optional filter predicate to apply to rows.
        partition_filters (Expression | None): Optional partition filter predicate to apply to partitions or files.
        columns (list[str] | None): Optional list of column names to project.
        limit (int | None): Optional limit on the number of rows to return.
    """

    filters: Expression | None = None
    partition_filters: Expression | None = None
    columns: list[str] | None = None
    limit: int | None = None

    @classmethod
    def _from_pypushdowns(cls, pushdowns: PyPushdowns) -> Pushdowns:
        return Pushdowns(
            filters=(Expression._from_pyexpr(pushdowns.filters) if pushdowns.filters else None),
            partition_filters=(
                Expression._from_pyexpr(pushdowns.partition_filters) if pushdowns.partition_filters else None
            ),
            columns=pushdowns.columns,
            limit=pushdowns.limit,
        )



from __future__ import annotations

import numpy as np
import pyarrow as pa
import pytest

import daft
from daft import col
from daft.datatype import DataType
from daft.expressions import Expression
from daft.expressions.testing import expr_structurally_equal
from daft.recordbatch import MicroPartition
from daft.series import Series
from daft.udf import udf
from tests.conftest import get_tests_daft_runner_name


def test_udf():
    table = MicroPartition.from_pydict({"a": ["foo", "bar", "baz"]})

    @udf(return_dtype=DataType.string())
    def repeat_n(data, n):
        return Series.from_pylist([d.as_py() * n for d in data.to_arrow()])

    expr = repeat_n(col("a"), 2)
    field = expr._to_field(table.schema())
    assert field.name == "a"
    assert field.dtype == DataType.string()

    result = table.eval_expression_list([expr])
    assert result.to_pydict() == {"a": ["foofoo", "barbar", "bazbaz"]}


@pytest.mark.parametrize("batch_size", [None, 1, 2, 3, 10])
@pytest.mark.parametrize("use_actor_pool", [False, True])
def test_class_udf(batch_size, use_actor_pool):
    df = daft.from_pydict({"a": ["foo", "bar", "baz"]})

    @udf(return_dtype=DataType.string(), batch_size=batch_size)
    class RepeatN:
        def __init__(self):
            self.n = 2

        def __call__(self, data):
            return Series.from_pylist([d.as_py() * self.n for d in data.to_arrow()])

    if use_actor_pool:
        RepeatN = RepeatN.with_concurrency(2)

    expr = RepeatN(col("a"))
    field = expr._to_field(df.schema())
    assert field.name == "a"
    assert field.dtype == DataType.string()

    result = df.select(expr)
    assert result.to_pydict() == {"a": ["foofoo", "barbar", "bazbaz"]}


@pytest.mark.parametrize("batch_size", [None, 1, 2, 3, 10])
@pytest.mark.parametrize("use_actor_pool", [False, True])
def test_class_udf_init_args(batch_size, use_actor_pool):
    df = daft.from_pydict({"a": ["foo", "bar", "baz"]})

    @udf(return_dtype=DataType.string(), batch_size=batch_size)
    class RepeatN:
        def __init__(self, initial_n: int = 2):
            self.n = initial_n

        def __call__(self, data):
            return Series.from_pylist([d.as_py() * self.n for d in data.to_arrow()])

    if use_actor_pool:
        RepeatN = RepeatN.with_concurrency(2)

    expr = RepeatN(col("a"))
    field = expr._to_field(df.schema())
    assert field.name == "a"
    assert field.dtype == DataType.string()
    result = df.select(expr)
    assert result.to_pydict() == {"a": ["foofoo", "barbar", "bazbaz"]}

    expr = RepeatN.with_init_args(initial_n=3)(col("a"))
    field = expr._to_field(df.schema())
    assert field.name == "a"
    assert field.dtype == DataType.string()
    result = df.select(expr)
    assert result.to_pydict() == {"a": ["foofoofoo", "barbarbar", "bazbazbaz"]}


@pytest.mark.parametrize("batch_size", [None, 1, 2, 3, 10])
@pytest.mark.parametrize("use_actor_pool", [False, True])
def test_class_udf_init_args_no_default(batch_size, use_actor_pool):
    df = daft.from_pydict({"a": ["foo", "bar", "baz"]})

    @udf(return_dtype=DataType.string(), batch_size=batch_size)
    class RepeatN:
        def __init__(self, initial_n):
            self.n = initial_n

        def __call__(self, data):
            return Series.from_pylist([d.as_py() * self.n for d in data.to_arrow()])

    if use_actor_pool:
        RepeatN = RepeatN.with_concurrency(2)

    with pytest.raises(ValueError, match="Cannot call class UDF without initialization arguments."):
        RepeatN(col("a"))

    expr = RepeatN.with_init_args(initial_n=2)(col("a"))
    field = expr._to_field(df.schema())
    assert field.name == "a"
    assert field.dtype == DataType.string()
    result = df.select(expr)
    assert result.to_pydict() == {"a": ["foofoo", "barbar", "bazbaz"]}


@pytest.mark.parametrize("use_actor_pool", [False, True])
def test_class_udf_init_args_bad_args(use_actor_pool):
    @udf(return_dtype=DataType.string())
    class RepeatN:
        def __init__(self, initial_n):
            self.n = initial_n

        def __call__(self, data):
            return Series.from_pylist([d.as_py() * self.n for d in data.to_arrow()])

    if use_actor_pool:
        RepeatN = RepeatN.with_concurrency(2)

    with pytest.raises(TypeError, match="missing a required argument: 'initial_n'"):
        RepeatN.with_init_args(wrong=5)


@pytest.mark.parametrize("concurrency", [1, 2, 4])
def test_actor_pool_udf_concurrency(concurrency):
    df = daft.from_pydict({"a": ["foo", "bar", "baz"]})

    @udf(return_dtype=DataType.string(), batch_size=1)
    class RepeatN:
        def __init__(self):
            self.n = 2

        def __call__(self, data):
            return Series.from_pylist([d.as_py() * self.n for d in data.to_arrow()])

    RepeatN = RepeatN.with_concurrency(concurrency)

    expr = RepeatN(col("a"))
    field = expr._to_field(df.schema())
    assert field.name == "a"
    assert field.dtype == DataType.string()

    result = df.select(expr)
    assert result.to_pydict() == {"a": ["foofoo", "barbar", "bazbaz"]}


def test_udf_kwargs():
    table = MicroPartition.from_pydict({"a": ["foo", "bar", "baz"]})

    @udf(return_dtype=DataType.string())
    def repeat_n(*, data=None, n=2):
        return Series.from_pylist([d.as_py() * n for d in data.to_arrow()])

    expr = repeat_n(data=col("a"))
    field = expr._to_field(table.schema())
    assert field.name == "a"
    assert field.dtype == DataType.string()

    result = table.eval_expression_list([expr])
    assert result.to_pydict() == {"a": ["foofoo", "barbar", "bazbaz"]}


@pytest.mark.parametrize("batch_size", [None, 1, 2, 3, 10])
def test_udf_tuples(batch_size):
    table = MicroPartition.from_pydict({"a": ["foo", "bar", "baz"]})

    @udf(return_dtype=DataType.string(), batch_size=batch_size)
    def repeat_n(data, tuple_data):
        n = tuple_data[0]
        return Series.from_pylist([d.as_py() * n for d in data.to_arrow()])

    expr = repeat_n(col("a"), (2,))
    field = expr._to_field(table.schema())
    assert field.name == "a"
    assert field.dtype == DataType.string()

    result = table.eval_expression_list([expr])
    assert result.to_pydict() == {"a": ["foofoo", "barbar", "bazbaz"]}


@pytest.mark.parametrize("container", [Series, list, np.ndarray])
@pytest.mark.parametrize("batch_size", [None, 1, 2, 3, 10])
def test_udf_return_containers(container, batch_size):
    table = MicroPartition.from_pydict({"a": ["foo", "bar", "baz"]})

    @udf(return_dtype=DataType.string(), batch_size=batch_size)
    def identity(data):
        if container is Series:
            return data
        elif container is list:
            return data.to_pylist()
        elif container is np.ndarray:
            return np.array(data.to_arrow())
        else:
            raise NotImplementedError(f"Test not implemented for container type: {container}")

    result = table.eval_expression_list([identity(col("a"))])
    assert result.to_pydict() == {"a": ["foo", "bar", "baz"]}


def test_udf_error():
    table = MicroPartition.from_pydict({"a": ["foo", "bar", "baz"]})

    @udf(return_dtype=DataType.string())
    def throw_value_err(x):
        raise ValueError("AN ERROR OCCURRED!")

    expr = throw_value_err(col("a"))

    with pytest.raises(ValueError) as exc_info:
        table.eval_expression_list([expr])

    assert str(exc_info.value).startswith("AN ERROR OCCURRED!\nUser-defined function ")
    assert str(exc_info.value).endswith("failed when executing on inputs:\n  - a (Utf8, length=3)")


@pytest.mark.parametrize("batch_size", [None, 1, 2, 3, 10])
@pytest.mark.parametrize("use_actor_pool", [False, True])
def test_no_args_udf_call(batch_size, use_actor_pool):
    @udf(return_dtype=DataType.int64(), batch_size=batch_size)
    def udf_no_args():
        pass

    if use_actor_pool:
        udf_no_args = udf_no_args.with_concurrency(2)

    assert isinstance(udf_no_args(), Expression)

    with pytest.raises(TypeError):
        udf_no_args("invalid")

    with pytest.raises(TypeError):
        udf_no_args(invalid="invalid")


@pytest.mark.parametrize("use_actor_pool", [False, True])
def test_full_udf_call(use_actor_pool):
    @udf(return_dtype=DataType.int64())
    def full_udf(e_arg, val, kwarg_val=None, kwarg_ex=None):
        pass

    if use_actor_pool:
        full_udf = full_udf.with_concurrency(2)

    assert isinstance(full_udf(col("x"), 1, kwarg_val=0, kwarg_ex=col("y")), Expression)

    with pytest.raises(TypeError):
        full_udf()


@pytest.mark.parametrize("use_actor_pool", [False, True])
def test_class_udf_initialization_error(use_actor_pool):
    df = daft.from_pydict({"a": ["foo", "bar", "baz"]})

    @udf(return_dtype=DataType.string())
    class IdentityWithInitError:
        def __init__(self):
            raise RuntimeError("UDF INIT ERROR")

        def __call__(self, data):
            return data

    if use_actor_pool:
        IdentityWithInitError = IdentityWithInitError.with_concurrency(1)

    expr = IdentityWithInitError(col("a"))
    if use_actor_pool:
        with pytest.raises(Exception):
            df.select(expr).collect()
    else:
        with pytest.raises(RuntimeError, match="UDF INIT ERROR"):
            df.select(expr).collect()


@pytest.mark.parametrize("use_actor_pool", [False, True])
def test_udf_equality(use_actor_pool):
    @udf(return_dtype=DataType.int64())
    def udf1(x):
        pass

    if use_actor_pool:
        udf1 = udf1.with_concurrency(2)

    assert expr_structurally_equal(udf1("x"), udf1("x"))
    assert not expr_structurally_equal(udf1("x"), udf1("y"))


@pytest.mark.parametrize("batch_size", [None, 1, 2, 3, 10])
def test_udf_return_tensor(batch_size):
    @udf(return_dtype=DataType.tensor(DataType.float64()), batch_size=batch_size)
    def np_udf(x):
        return [np.ones((3, 3)) * i for i in x.to_pylist()]

    expr = np_udf(col("x"))
    table = MicroPartition.from_pydict({"x": [0, 1, 2]})
    result = table.eval_expression_list([expr])
    assert len(result.to_pydict()["x"]) == 3
    for i in range(3):
        np.testing.assert_array_equal(result.to_pydict()["x"][i], np.ones((3, 3)) * i)


@pytest.mark.parametrize("batch_size", [None, 1, 2, 3, 10])
@pytest.mark.parametrize("use_actor_pool", [False, True])
def test_udf_return_embedding(batch_size, use_actor_pool):
    # Create test data
    table = MicroPartition.from_pydict({"x": [0, 1, 2]})

    # Define UDF that returns an embedding
    @udf(return_dtype=DataType.embedding(DataType.float32(), 2), batch_size=batch_size)
    def embedding_udf(x):
        # Create a 2D embedding for each input value
        return [np.array([i, i + 1], dtype=np.float32) for i in x.to_pylist()]

    # Apply the UDF
    if use_actor_pool:
        embedding_udf = embedding_udf.with_concurrency(2)

    expr = embedding_udf(col("x"))
    result = table.eval_expression_list([expr])

    # Verify results
    embeddings = result.to_pydict()["x"]
    assert len(embeddings) == 3

    # Check each embedding vector
    np.testing.assert_array_equal(embeddings[0], np.array([0, 1], dtype=np.float32))
    np.testing.assert_array_equal(embeddings[1], np.array([1, 2], dtype=np.float32))
    np.testing.assert_array_equal(embeddings[2], np.array([2, 3], dtype=np.float32))


@pytest.mark.skip(
    reason="[RUST-INT][UDF] repr is very naive at the moment py_udf(...exprs), we should fix to show all parameters and use the function name"
)
def test_udf_repr():
    @udf(return_dtype=DataType.int64(), expr_inputs=["x"])
    def single_arg_udf(x, y, z=10):
        pass

    assert single_arg_udf(col("x"), "y", z=20).__repr__() == "@udf[single_arg_udf](col('x'), 'y', z=20)"


@pytest.mark.parametrize("batch_size", [None, 1, 2, 3, 10])
def test_udf_arbitrary_number_of_args(batch_size):
    table = MicroPartition.from_pydict({"a": [1, 2, 3], "b": [1, 2, 3], "c": [1, 2, 3]})

    @udf(return_dtype=DataType.int64(), batch_size=batch_size)
    def add_cols_elementwise(*args):
        return Series.from_pylist([sum(x) for x in zip(*[arg.to_pylist() for arg in args])])

    expr = add_cols_elementwise(col("a"), col("b"), col("c"))
    field = expr._to_field(table.schema())
    assert field.name == "a"
    assert field.dtype == DataType.int64()

    result = table.eval_expression_list([expr])
    assert result.to_pydict() == {"a": [3, 6, 9]}


@pytest.mark.parametrize("batch_size", [None, 1, 2, 3, 10])
def test_udf_arbitrary_number_of_kwargs(batch_size):
    table = MicroPartition.from_pydict({"a": [1, 2, 3], "b": [1, 2, 3], "c": [1, 2, 3]})

    @udf(return_dtype=DataType.string(), batch_size=batch_size)
    def repeat_kwargs(**kwargs):
        data = {k: v.to_pylist() for k, v in kwargs.items()}
        length = len(data[next(iter(data.keys()))])
        return Series.from_pylist(["".join([key * data[key][i] for key in data]) for i in range(length)])

    expr = repeat_kwargs(a=col("a"), b=col("b"), c=col("c"))
    field = expr._to_field(table.schema())
    assert field.name == "a"
    assert field.dtype == DataType.string()

    result = table.eval_expression_list([expr])
    assert result.to_pydict() == {"a": ["abc", "aabbcc", "aaabbbccc"]}


@pytest.mark.parametrize("batch_size", [None, 1, 2, 3, 10])
def test_udf_arbitrary_number_of_args_with_kwargs(batch_size):
    table = MicroPartition.from_pydict({"a": [1, 2, 3], "b": [1, 2, 3], "c": [1, 2, 3]})

    @udf(return_dtype=DataType.int64(), batch_size=batch_size)
    def add_cols_elementwise(*args, multiplier: float):
        return Series.from_pylist([multiplier * sum(x) for x in zip(*[arg.to_pylist() for arg in args])])

    expr = add_cols_elementwise(col("a"), col("b"), col("c"), multiplier=2)
    field = expr._to_field(table.schema())
    assert field.name == "a"
    assert field.dtype == DataType.int64()

    result = table.eval_expression_list([expr])
    assert result.to_pydict() == {"a": [6, 12, 18]}


@pytest.mark.parametrize("batch_size", [None, 1, 2, 3, 10])
def test_udf_return_pyarrow(batch_size):
    table = MicroPartition.from_pydict({"a": [1, 2, 3]})

    @udf(return_dtype=DataType.int64(), batch_size=batch_size)
    def add_1(data):
        return pa.compute.add(data.to_arrow(), 1)

    result = table.eval_expression_list([add_1(col("a"))])
    assert result.to_pydict() == {"a": [2, 3, 4]}


@pytest.mark.parametrize("batch_size", [1, 2, 3, 4, 7, 8, 1000])
def test_udf_batch_size(batch_size):
    table = MicroPartition.from_pydict({"a": [1, 2, 3, 4, 5, 6, 7]})

    @udf(return_dtype=DataType.int64(), batch_size=batch_size)
    def add_1(data):
        lt = data.to_pylist()
        assert len(lt) <= batch_size
        return [x + 1 for x in lt]

    result = table.eval_expression_list([add_1(col("a"))])
    assert result.to_pydict() == {"a": [2, 3, 4, 5, 6, 7, 8]}


@pytest.mark.parametrize("batch_size", [1, 2, 3, 4, 7, 8, 1000])
def test_udf_batch_size_override(batch_size):
    table = MicroPartition.from_pydict({"a": [1, 2, 3, 4, 5, 6, 7]})

    @udf(return_dtype=DataType.int64())
    def add_1(data):
        lt = data.to_pylist()
        assert len(lt) <= batch_size
        return [x + 1 for x in lt]

    result = table.eval_expression_list([add_1.override_options(batch_size=batch_size)(col("a"))])
    assert result.to_pydict() == {"a": [2, 3, 4, 5, 6, 7, 8]}


def test_udf_invalid_batch_sizes():
    table = MicroPartition.from_pydict({"a": [1, 2, 3, 4, 5, 6, 7]})

    @udf(return_dtype=DataType.int64())
    def noop(data):
        return data

    with pytest.raises(ValueError, match="batch size must be positive"):
        table.eval_expression_list([noop.override_options(batch_size=0)(col("a"))])

    with pytest.raises(OverflowError):
        table.eval_expression_list([noop.override_options(batch_size=-1)(col("a"))])


@pytest.mark.parametrize("batch_size", [None, 1, 2])
@pytest.mark.parametrize("use_actor_pool", [False, True])
def test_udf_empty(batch_size, use_actor_pool):
    df = daft.from_pydict({"a": []})

    @udf(return_dtype=DataType.int64(), batch_size=batch_size)
    def identity(data):
        return data

    if use_actor_pool:
        identity = identity.with_concurrency(2)

    result = df.select(identity(col("a")))
    assert result.to_pydict() == {"a": []}


@pytest.mark.skipif(
    get_tests_daft_runner_name() not in {"native", "ray"},
    reason="requires Native or Ray Runner to be in use",
)
@pytest.mark.parametrize("use_actor_pool", [False, True])
def test_udf_with_error(use_actor_pool):
    import re

    df = daft.from_pydict({"a": [1, 2, 3], "b": ["foo", "bar", "baz"]})

    @udf(return_dtype=DataType.int64())
    def fail_hard(a, b):
        raise ValueError("AN ERROR OCCURRED!")

    if use_actor_pool:
        fail_hard = fail_hard.with_concurrency(2)

    with pytest.raises(Exception) as exc_info:
        df.select(fail_hard(col("a"), col("b"))).collect()

    pattern = (
        r"AN ERROR OCCURRED!\n"
        r"User-defined function `<function test_udf_with_error\.<locals>\.fail_hard at 0x[0-9a-f]+>` "
        r"failed when executing on inputs:\s*"
        r"- a \(Int64, length=3\)\s*"
        r"- b \(Utf8, length=3\)$"
    )
    assert re.search(pattern, str(exc_info.value)), f"String doesn't end with expected pattern: {exc_info.value!s}"



from __future__ import annotations

import pathlib
from itertools import chain
from typing import TYPE_CHECKING, Any, Literal

import lance

from daft.context import get_context
from daft.datatype import DataType
from daft.io import DataSink
from daft.io.sink import WriteResult
from daft.recordbatch import MicroPartition
from daft.schema import Schema

if TYPE_CHECKING:
    from collections.abc import Iterator
    from types import ModuleType

    from daft.daft import IOConfig


class LanceDataSink(DataSink[list[lance.FragmentMetadata]]):
    """WriteSink for writing data to a Lance dataset."""

    def _import_lance(self) -> ModuleType:
        try:
            import lance

            return lance
        except ImportError:
            raise ImportError("lance is not installed. Please install lance using `pip install daft[lance]`")

    def __init__(
        self,
        uri: str | pathlib.Path,
        schema: Schema,
        mode: Literal["create", "append", "overwrite"],
        io_config: IOConfig | None = None,
        **kwargs: Any,
    ) -> None:
        from daft.dependencies import pa
        from daft.io.object_store_options import io_config_to_storage_options

        lance = self._import_lance()

        if not isinstance(uri, (str, pathlib.Path)):
            raise TypeError(f"Expected URI to be str or pathlib.Path, got {type(uri)}")
        self._table_uri = str(uri)
        self._mode = mode
        self._io_config = get_context().daft_planning_config.default_io_config if io_config is None else io_config
        self._kwargs = kwargs

        self._storage_options = io_config_to_storage_options(self._io_config, self._table_uri)

        self._pyarrow_schema = pa.schema((f.name, f.dtype.to_arrow_dtype()) for f in schema)

        try:
            table = lance.dataset(self._table_uri, storage_options=self._storage_options)

        except ValueError:
            table = None

        self._version = 0
        if table:
            table_schema = table.schema
            self._version = table.latest_version
            if self._pyarrow_schema != table_schema and not (self._mode == "overwrite"):
                raise ValueError(
                    "Schema of data does not match table schema\n"
                    f"Data schema:\n{self._pyarrow_schema}\nTable Schema:\n{table_schema}"
                )

        self._schema = Schema._from_field_name_and_types(
            [
                ("num_fragments", DataType.int64()),
                ("num_deleted_rows", DataType.int64()),
                ("num_small_files", DataType.int64()),
                ("version", DataType.int64()),
            ]
        )

    def schema(self) -> Schema:
        return self._schema

    def write(self, micropartitions: Iterator[MicroPartition]) -> Iterator[WriteResult[list[lance.FragmentMetadata]]]:
        """Writes fragments from the given micropartitions."""
        lance = self._import_lance()

        for micropartition in micropartitions:
            arrow_table = micropartition.to_arrow()
            bytes_written = arrow_table.nbytes
            rows_written = arrow_table.num_rows

            fragments = lance.fragment.write_fragments(
                arrow_table,
                dataset_uri=self._table_uri,
                mode=self._mode,
                storage_options=self._storage_options,
                **self._kwargs,
            )
            yield WriteResult(
                result=fragments,
                bytes_written=bytes_written,
                rows_written=rows_written,
            )

    def finalize(self, write_results: list[WriteResult[list[lance.FragmentMetadata]]]) -> MicroPartition:
        """Commits the fragments to the Lance dataset. Returns a DataFrame with the stats of the dataset."""
        from daft.dependencies import pa

        lance = self._import_lance()

        fragments = list(chain.from_iterable(write_result.result for write_result in write_results))

        if self._mode == "create" or self._mode == "overwrite":
            operation = lance.LanceOperation.Overwrite(self._pyarrow_schema, fragments)
        elif self._mode == "append":
            operation = lance.LanceOperation.Append(fragments)

        dataset = lance.LanceDataset.commit(
            self._table_uri, operation, read_version=self._version, storage_options=self._storage_options
        )
        stats = dataset.stats.dataset_stats()

        tbl = MicroPartition.from_pydict(
            {
                "num_fragments": pa.array([stats["num_fragments"]], type=pa.int64()),
                "num_deleted_rows": pa.array([stats["num_deleted_rows"]], type=pa.int64()),
                "num_small_files": pa.array([stats["num_small_files"]], type=pa.int64()),
                "version": pa.array([dataset.version], type=pa.int64()),
            }
        )
        return tbl



from __future__ import annotations

import os
import zipfile
from pathlib import Path

import pytest


def _extract_testing_table(zip_file_path, target_path, table_name) -> str:
    with zipfile.ZipFile(zip_file_path, "r") as zip_ref:
        zip_ref.extractall(target_path)
    return os.path.join(target_path, table_name)


@pytest.fixture(
    params=[
        "v6_complexkeygen_hivestyle",
        "v6_nonpartitioned",
        "v6_simplekeygen_nonhivestyle",
        "v6_simplekeygen_hivestyle_no_metafields",
        "v6_timebasedkeygen_nonhivestyle",
    ]
)
def get_testing_table_for_supported_cases(request, tmp_path) -> str:
    table_name = request.param
    zip_file_path = Path(__file__).parent.joinpath("data", f"{table_name}.zip")
    return _extract_testing_table(zip_file_path, tmp_path, table_name)


@pytest.fixture(
    params=[
        "v6_empty",
    ]
)
def get_empty_table(request, tmp_path) -> str:
    table_name = request.param
    zip_file_path = Path(__file__).parent.joinpath("data", f"{table_name}.zip")
    return _extract_testing_table(zip_file_path, tmp_path, table_name)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Callable
from urllib.parse import urlparse

from daft.dependencies import pa
from daft.logical.schema import Schema

if TYPE_CHECKING:
    from sqlalchemy.engine import Connection


logger = logging.getLogger(__name__)


class SQLConnection:
    def __init__(self, conn: str | Callable[[], Connection], driver: str, dialect: str, url: str) -> None:
        self.conn = conn
        self.dialect = dialect
        self.driver = driver
        self.url = url

    def __repr__(self) -> str:
        return f"SQLConnection(conn={self.conn})"

    @classmethod
    def from_url(cls, url: str) -> SQLConnection:
        scheme = urlparse(url).scheme.strip().lower()
        if "+" in scheme:
            dialect, driver = scheme.split("+")
        else:
            dialect, driver = scheme, ""
        return cls(url, driver, dialect, url)

    @classmethod
    def from_connection_factory(cls, conn_factory: Callable[[], Connection]) -> SQLConnection:
        from sqlalchemy.engine import Connection

        try:
            with conn_factory() as connection:
                if not isinstance(connection, Connection):
                    raise ValueError(
                        f"Connection factory must return a SQLAlchemy connection object, got: {type(connection)}"
                    )
                dialect = connection.engine.dialect.name
                driver = connection.engine.driver
                url = connection.engine.url.render_as_string()
            return cls(conn_factory, driver, dialect, url)
        except Exception as e:
            raise ValueError(f"Unexpected error while calling the connection factory: {e}") from e

    def read_schema(self, sql: str, infer_schema_length: int) -> Schema:
        if self._should_use_connectorx():
            sql = self.construct_sql_query(sql, limit=0)
        else:
            sql = self.construct_sql_query(sql, limit=infer_schema_length)
        table = self.execute_sql_query(sql)
        schema = Schema.from_pyarrow_schema(table.schema)
        return schema

    def construct_sql_query(
        self,
        sql: str,
        projection: list[str] | None = None,
        predicate: str | None = None,
        limit: int | None = None,
        partition_bounds: tuple[str, str] | None = None,
    ) -> str:
        # If all options are None, just return the original sql
        if projection is None and predicate is None and limit is None and partition_bounds is None:
            return sql

        import sqlglot

        target_dialect = self.dialect
        # sqlglot does not support "postgresql" dialect, it only supports "postgres"
        if target_dialect == "postgresql":
            target_dialect = "postgres"
        # sqlglot does not recognize "mssql" as a dialect, it instead recognizes "tsql", which is the SQL dialect for Microsoft SQL Server
        elif target_dialect == "mssql":
            target_dialect = "tsql"

        if not any(target_dialect == supported_dialect.value for supported_dialect in sqlglot.Dialects):
            raise ValueError(
                f"Unsupported dialect: {target_dialect}, please refer to the documentation for supported dialects."
            )

        query = sqlglot.subquery(sql, "subquery", dialect=target_dialect)

        if projection is not None:
            query = query.select(*projection)
        else:
            query = query.select("*")

        if predicate is not None:
            query = query.where(predicate)

        if partition_bounds is not None:
            query = query.where(partition_bounds[0]).where(partition_bounds[1])

        if limit is not None:
            query = query.limit(limit)

        return query.sql(dialect=target_dialect)

    def _should_use_connectorx(self) -> bool:
        # Supported DBs extracted from here https://github.com/sfu-db/connector-x/tree/7b3147436b7e20b96691348143d605e2249d6119?tab=readme-ov-file#sources
        connectorx_supported_dbs = {
            "postgres",
            "postgresql",
            "mysql",
            "mssql",
            "oracle",
            "bigquery",
            "sqlite",
            "clickhouse",
            "redshift",
        }

        if isinstance(self.conn, str):
            if self.dialect in connectorx_supported_dbs and self.driver == "":
                return True
        return False

    def execute_sql_query(self, sql: str) -> pa.Table:
        if self._should_use_connectorx():
            return self._execute_sql_query_with_connectorx(sql)
        else:
            return self._execute_sql_query_with_sqlalchemy(sql)

    def _execute_sql_query_with_connectorx(self, sql: str) -> pa.Table:
        import connectorx as cx

        assert isinstance(self.conn, str)
        logger.info("Using connectorx to execute sql: %s", sql)
        try:
            table = cx.read_sql(conn=self.conn, query=sql, return_type="arrow")
            return table
        except Exception as e:
            raise RuntimeError(f"Failed to execute sql: {sql} with url: {self.conn}, error: {e}") from e

    def _execute_sql_query_with_sqlalchemy(self, sql: str) -> pa.Table:
        from sqlalchemy import create_engine, text

        logger.info("Using sqlalchemy to execute sql: %s", sql)
        try:
            if isinstance(self.conn, str):
                with create_engine(self.conn).connect() as connection:
                    result = connection.execute(text(sql))
                    rows = result.fetchall()
            else:
                with self.conn() as connection:
                    result = connection.execute(text(sql))
                    rows = result.fetchall()

            pydict = {column_name: [row[i] for row in rows] for i, column_name in enumerate(result.keys())}
            # TODO: Use type codes from cursor description to create pyarrow schema
            return pa.Table.from_pydict(pydict)
        except Exception as e:
            connection_str = self.conn if isinstance(self.conn, str) else self.conn.__name__
            raise RuntimeError(f"Failed to execute sql: {sql} from connection: {connection_str}, error: {e}") from e



"""WARNING! These APIs are internal; please use Catalog.from_unity() and Table.from_unity()."""

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any

from unitycatalog import NotFoundError as UnityNotFoundError

from daft.catalog import Catalog, Identifier, NotFoundError, Properties, Schema, Table
from daft.io._deltalake import read_deltalake
from daft.unity_catalog import UnityCatalog as InnerCatalog  # noqa: TID253
from daft.unity_catalog import UnityCatalogTable as InnerTable  # noqa: TID253

if TYPE_CHECKING:
    from daft.dataframe import DataFrame


class UnityCatalog(Catalog):
    _inner: InnerCatalog

    def __init__(self, unity_catalog: InnerCatalog):
        """DEPRECATED: Please use `Catalog.from_unity`; version 0.5.0!"""
        warnings.warn(
            "This is deprecated and will be removed in daft >= 0.5.0, please prefer using `Catalog.from_unity` instead; version 0.5.0!",
            category=DeprecationWarning,
        )
        self._inner = unity_catalog

    @property
    def name(self) -> str:
        # TODO feat: add names to unity catalogs
        return "unity"

    @staticmethod
    def _from_obj(obj: object) -> UnityCatalog:
        """Returns an UnityCatalog instance if the given object can be adapted so."""
        if isinstance(obj, InnerCatalog):
            return UnityCatalog(obj)
        raise ValueError(f"Unsupported unity catalog type: {type(obj)}")

    ###
    # create_*
    ###

    def _create_namespace(self, identifier: Identifier) -> None:
        raise NotImplementedError("Unity create_namespace not yet supported.")

    def _create_table(
        self,
        identifier: Identifier,
        source: Schema,
        properties: Properties | None = None,
    ) -> Table:
        raise NotImplementedError("Unity create_table not yet supported.")

    ###
    # drop_*
    ###

    def _drop_namespace(self, identifier: Identifier) -> None:
        raise NotImplementedError("Unity drop_namespace not yet supported.")

    def _drop_table(self, identifier: Identifier) -> None:
        raise NotImplementedError("Unity drop_table not yet supported.")

    ###
    # get_*
    ###

    def _get_table(self, ident: Identifier) -> UnityTable:
        try:
            return UnityTable(self._inner.load_table(str(ident)))
        except UnityNotFoundError:
            raise NotFoundError(f"Table {ident} not found!")

    ###
    # list_.*
    ###

    def _list_namespaces(self, pattern: str | None = None) -> list[Identifier]:
        raise NotImplementedError("Unity list_namespaces not yet supported.")

    def _list_tables(self, pattern: str | None = None) -> list[Identifier]:
        if pattern is None or pattern == "":
            return [
                Identifier.from_str(tbl)
                for cat in self._inner.list_catalogs()
                for schema in self._inner.list_schemas(cat)
                for tbl in self._inner.list_tables(schema)
            ]
        num_namespaces = pattern.count(".")
        if num_namespaces == 0:
            catalog_name = pattern
            return [
                Identifier.from_str(tbl)
                for schema in self._inner.list_schemas(catalog_name)
                for tbl in self._inner.list_tables(schema)
            ]
        elif num_namespaces == 1:
            schema_name = pattern
            return [Identifier.from_str(tbl) for tbl in self._inner.list_tables(schema_name)]
        else:
            raise ValueError(
                f"Unrecognized catalog name or schema name, expected a '.'-separated namespace but received: {pattern}"
            )

    ###
    # has_.*
    ###
    def _has_namespace(self, ident: Identifier) -> bool:
        raise NotImplementedError("Unity has_namespace not yet supported.")

    def _has_table(self, ident: Identifier) -> bool:
        try:
            self._inner.load_table(str(ident))
            return True
        except UnityNotFoundError:
            return False


class UnityTable(Table):
    _inner: InnerTable

    _read_options = {"version"}
    _write_options = {
        "schema_mode",
        "partition_col",
        "description",
        "configuration",
        "custom_metadata",
        "dynamo_table_name",
        "allow_unsafe_rename",
    }

    def __init__(self, unity_table: InnerTable):
        """DEPRECATED: Please use `Table.from_unity`; version 0.5.0!"""
        warnings.warn(
            "This is deprecated and will be removed in daft >= 0.5.0, please prefer using `Table.from_unity` instead; version 0.5.0!",
            category=DeprecationWarning,
        )
        self._inner = unity_table

    @property
    def name(self) -> str:
        return self._inner.table_info.name

    def schema(self) -> Schema:
        return self.read().schema()

    @staticmethod
    def _from_obj(obj: object) -> UnityTable:
        """Returns a UnityTable if the given object can be adapted so."""
        if isinstance(obj, InnerTable):
            t = UnityTable.__new__(UnityTable)
            t._inner = obj
            return t
        raise ValueError(f"Unsupported unity table type: {type(obj)}")

    ###
    # read methods
    ###

    def read(self, **options: Any) -> DataFrame:
        Table._validate_options("Unity read", options, UnityTable._read_options)

        return read_deltalake(self._inner, version=options.get("version"))

    ###
    # write methods
    ###

    def append(self, df: DataFrame, **options: Any) -> None:
        self._validate_options("Unity write", options, UnityTable._write_options)

        df.write_deltalake(self._inner, mode="append", **options)

    def overwrite(self, df: DataFrame, **options: Any) -> None:
        self._validate_options("Unity write", options, UnityTable._write_options)

        df.write_deltalake(self._inner, mode="overwrite", **options)
